"""Definition of Maxwell spaces."""

import numpy as _np
import numba as _numba


def _is_screen(grid):
    """Check if there is an edge only adjacent to one triangle."""
    for e in range(grid.edges.shape[1]):
        if len([j for i in grid.element_edges for j in i if j == e]) < 2:
            return True
    return False


def rwg0_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=False,
    truncate_at_segment_edge=True,
):
    """Define a space of RWG functions of order 0."""
    from .space import SpaceBuilder, _process_segments
    from bempp_cl.api.utils.helpers import serialise_list_of_lists

    support, normal_multipliers = _process_segments(grid, support_elements, segments, swapped_normals)

    edge_neighbors, edge_neighbors_ptr = serialise_list_of_lists(grid.edge_neighbors)

    (
        global_dof_count,
        support,
        local2global,
        local_multipliers,
    ) = _compute_rwg0_space_data(
        support,
        edge_neighbors,
        edge_neighbors_ptr,
        grid.element_edges,
        grid.number_of_elements,
        grid.number_of_edges,
        include_boundary_dofs,
        truncate_at_segment_edge,
    )

    return (
        SpaceBuilder(grid)
        .set_codomain_dimension(3)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(False)
        .set_shapeset("rwg0")
        .set_identifier("rwg0")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_barycentric_representation(rwg0_barycentric_function_space)
        .set_numba_evaluator(_numba_rwg0_evaluate)
        .build()
    )


def rwg0_barycentric_function_space(coarse_space):
    """Define a space of RWG functions of order 0 over a barycentric grid."""
    from .space import SpaceBuilder
    from scipy.sparse import coo_matrix

    number_of_support_elements = coarse_space.number_of_support_elements
    bary_grid_number_of_elements = 6 * coarse_space.grid.number_of_elements

    bary_support_elements = 6 * _np.repeat(coarse_space.support_elements, 6) + _np.tile(
        _np.arange(6), number_of_support_elements
    )

    bary_support_size = len(bary_support_elements)

    support = _np.zeros(6 * coarse_space.grid.number_of_elements, dtype=_np.bool_)
    support[bary_support_elements] = True

    normal_multipliers = _np.repeat(coarse_space.normal_multipliers, 6)

    local_coords = _np.array([[0, 0], [0.5, 0], [1, 0], [0.5, 0.5], [0, 1], [0, 0.5], [1.0 / 3, 1.0 / 3]]).T

    coeffs = (
        _np.array(
            [
                [1, -1.0 / 3, 0],
                [-1.0 / 3, 1, 0],
                [0, 1.0 / 3, -1.0 / 6],
                [0, 0, 1.0 / 6],
                [0, 0, 1.0 / 6],
                [1.0 / 3, 0, -1.0 / 6],
            ]
        ),
        _np.array(
            [
                [0, 1.0 / 3, -1.0 / 6],
                [0, 0, 1.0 / 6],
                [0, 0, 1.0 / 6],
                [1.0 / 3, 0, -1.0 / 6],
                [1, -1.0 / 3, 0],
                [-1.0 / 3, 1, 0],
            ]
        ),
        _np.array(
            [
                [0, 0, 1.0 / 6],
                [1.0 / 3, 0, -1.0 / 6],
                [1, -1.0 / 3, 0],
                [-1.0 / 3, 1, 0],
                [0, 1.0 / 3, -1.0 / 6],
                [0, 0, 1.0 / 6],
            ]
        ),
    )

    coarse_dofs, bary_dofs, values = generate_rwg0_map(
        coarse_space.grid.data(), coarse_space.support_elements, local_coords, coeffs
    )

    local2global = _np.zeros((bary_grid_number_of_elements, 3), dtype="uint32")
    local_multipliers = _np.zeros((bary_grid_number_of_elements, 3), dtype="uint32")

    local2global[support] = _np.arange(3 * bary_support_size).reshape(bary_support_size, 3)

    local_multipliers[support] = 1

    transform = coo_matrix(
        (values, (bary_dofs, coarse_dofs)),
        shape=(3 * bary_support_size, 3 * number_of_support_elements),
        dtype=_np.float64,
    ).tocsr()

    dof_transformation = transform @ coarse_space.map_to_localised_space

    return (
        SpaceBuilder(coarse_space.grid.barycentric_refinement)
        .set_codomain_dimension(3)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("rwg0")
        .set_identifier("rwg0")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_dof_transformation(dof_transformation)
        .set_numba_evaluator(_numba_rwg0_evaluate)
        .build()
    )


def snc0_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=False,
    truncate_at_segment_edge=True,
):
    """Define a space of SNC functions of order 0."""
    from .space import SpaceBuilder, _process_segments
    from bempp_cl.api.utils.helpers import serialise_list_of_lists

    support, normal_multipliers = _process_segments(grid, support_elements, segments, swapped_normals)

    edge_neighbors, edge_neighbors_ptr = serialise_list_of_lists(grid.edge_neighbors)

    (
        global_dof_count,
        support,
        local2global,
        local_multipliers,
    ) = _compute_rwg0_space_data(
        support,
        edge_neighbors,
        edge_neighbors_ptr,
        grid.element_edges,
        grid.number_of_elements,
        grid.number_of_edges,
        include_boundary_dofs,
        truncate_at_segment_edge,
    )

    return (
        SpaceBuilder(grid)
        .set_codomain_dimension(3)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(False)
        .set_shapeset("snc0")
        .set_identifier("snc0")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_barycentric_representation(snc0_barycentric_function_space)
        .set_numba_evaluator(_numba_snc0_evaluate)
        .set_numba_surface_curl(_numba_snc0_surface_curl)
        .build()
    )


def snc0_barycentric_function_space(coarse_space):
    """Define a space of SNC functions of order 0 over a barycentric grid."""
    from .space import SpaceBuilder
    from scipy.sparse import coo_matrix

    number_of_support_elements = coarse_space.number_of_support_elements
    bary_grid_number_of_elements = 6 * coarse_space.grid.number_of_elements

    bary_support_elements = 6 * _np.repeat(coarse_space.support_elements, 6) + _np.tile(
        _np.arange(6), number_of_support_elements
    )

    bary_support_size = len(bary_support_elements)

    support = _np.zeros(6 * coarse_space.grid.number_of_elements, dtype=_np.bool_)
    support[bary_support_elements] = True

    normal_multipliers = _np.repeat(coarse_space.normal_multipliers, 6)

    local_coords = _np.array([[0, 0], [0.5, 0], [1, 0], [0.5, 0.5], [0, 1], [0, 0.5], [1.0 / 3, 1.0 / 3]]).T

    coeffs = (
        _np.array(
            [
                [1, -1.0 / 3, 0],
                [-1.0 / 3, 1, 0],
                [0, 1.0 / 3, -1.0 / 6],
                [0, 0, 1.0 / 6],
                [0, 0, 1.0 / 6],
                [1.0 / 3, 0, -1.0 / 6],
            ]
        ),
        _np.array(
            [
                [0, 1.0 / 3, -1.0 / 6],
                [0, 0, 1.0 / 6],
                [0, 0, 1.0 / 6],
                [1.0 / 3, 0, -1.0 / 6],
                [1, -1.0 / 3, 0],
                [-1.0 / 3, 1, 0],
            ]
        ),
        _np.array(
            [
                [0, 0, 1.0 / 6],
                [1.0 / 3, 0, -1.0 / 6],
                [1, -1.0 / 3, 0],
                [-1.0 / 3, 1, 0],
                [0, 1.0 / 3, -1.0 / 6],
                [0, 0, 1.0 / 6],
            ]
        ),
    )

    coarse_dofs, bary_dofs, values = generate_rwg0_map(
        coarse_space.grid.data(), coarse_space.support_elements, local_coords, coeffs
    )

    local2global = _np.zeros((bary_grid_number_of_elements, 3), dtype="uint32")
    local_multipliers = _np.zeros((bary_grid_number_of_elements, 3), dtype="uint32")

    local2global[support] = _np.arange(3 * bary_support_size).reshape(bary_support_size, 3)

    local_multipliers[support] = 1

    transform = coo_matrix(
        (values, (bary_dofs, coarse_dofs)),
        shape=(3 * bary_support_size, 3 * number_of_support_elements),
        dtype=_np.float64,
    ).tocsr()

    dof_transformation = transform @ coarse_space.map_to_localised_space

    return (
        SpaceBuilder(coarse_space.grid.barycentric_refinement)
        .set_codomain_dimension(3)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("rwg0")
        .set_identifier("snc0")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_dof_transformation(dof_transformation)
        .set_numba_evaluator(_numba_snc0_evaluate)
        .build()
    )


def bc_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=False,
    truncate_at_segment_edge=True,
):
    """Define a space of BC functions."""
    from .space import SpaceBuilder

    if _is_screen(grid) and include_boundary_dofs:
        raise ValueError("Boundary dofs inclusion is not implemented for BC yet")

    bary_grid = grid.barycentric_refinement

    coarse_space = rwg0_function_space(
        grid,
        support_elements,
        segments,
        swapped_normals,
        include_boundary_dofs=include_boundary_dofs,
        truncate_at_segment_edge=truncate_at_segment_edge,
    )

    (
        dof_transformation,
        support,
        normal_multipliers,
        local2global,
        local_multipliers,
    ) = _compute_bc_space_data(grid, bary_grid, coarse_space, truncate_at_segment_edge, swapped_normals)

    return (
        SpaceBuilder(bary_grid)
        .set_codomain_dimension(3)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("rwg0")
        .set_identifier("rwg0")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_dof_transformation(dof_transformation)
        .set_numba_evaluator(_numba_rwg0_evaluate)
        .build()
    )


def rbc_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=False,
    truncate_at_segment_edge=True,
):
    """Define a space of RBC functions."""
    from .space import SpaceBuilder

    if _is_screen(grid) and include_boundary_dofs:
        raise ValueError("Boundary dofs inclusion is not implemented for RBC yet")

    bary_grid = grid.barycentric_refinement

    coarse_space = rwg0_function_space(
        grid,
        support_elements,
        segments,
        swapped_normals,
        include_boundary_dofs=include_boundary_dofs,
        truncate_at_segment_edge=truncate_at_segment_edge,
    )

    (
        dof_transformation,
        support,
        normal_multipliers,
        local2global,
        local_multipliers,
    ) = _compute_bc_space_data(grid, bary_grid, coarse_space, truncate_at_segment_edge, swapped_normals)

    return (
        SpaceBuilder(bary_grid)
        .set_codomain_dimension(3)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("rwg0")
        .set_identifier("snc0")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_dof_transformation(dof_transformation)
        .set_numba_evaluator(_numba_snc0_evaluate)
        .build()
    )


def _compute_bc_space_data(grid, bary_grid, coarse_space, truncate_at_segment_edge, swapped_normals):
    """Generate the BC map."""
    from bempp_cl.api.grid.grid import (
        _get_barycentric_support,
        _get_bary_coefficients,
        _get_coefficients_reference_edge,
        _get_data_multipliers,
        _get_barycentric_edges_associated_to_vertex,
    )

    from scipy.sparse import coo_matrix

    support, bary_support_size, bary_support_elements = _get_barycentric_support(
        truncate_at_segment_edge, grid, bary_grid, coarse_space
    )

    bary_vertex_to_edge, local2global, edge_lengths, normal_multipliers, local_multipliers = _get_data_multipliers(
        support, bary_grid, bary_support_elements, swapped_normals, coarse_space, bary_support_size
    )

    coarse_dofs = []
    bary_dofs = []
    values = []

    for global_dof_index in range(coarse_space.global_dof_count):
        # Local numbering of the edges consituting the support.
        local_dofs = coarse_space.global2local[global_dof_index]

        # Edge within the coarse support (reference edge).
        edge_index = grid.data().element_edges[local_dofs[0][1], local_dofs[0][0]]

        # Coarse elements associated to a reference edge
        neighbors = grid.edge_neighbors[edge_index]

        # Find upper and lower coarse elements over reference edge.
        other = neighbors[1] if local_dofs[0][0] == neighbors[0] else neighbors[0]
        if coarse_space.local_multipliers[local_dofs[0][0], local_dofs[0][1]] > 0:
            lower = local_dofs[0][0]
            upper = other
        else:
            lower = other
            upper = local_dofs[0][0]

        # For the BC function, pick the left and right vertex to which the
        # barycentric elements are associated to.
        vertex1, vertex2 = grid.data().edges[:, edge_index]

        # Define the right and left vertex depending on which is the upper
        # and lower element over the reference edge.
        for local_index, vertex_index in enumerate(grid.data().elements[:, upper]):
            if vertex_index == vertex1:
                break

        if vertex2 == grid.data().elements[(local_index - 1) % 3, upper]:
            vertex1, vertex2 = vertex2, vertex1

        # Get the local coarse vertex indices (0, 1 or 2) of vertex1 and vertex2 in upper and
        # lower coarse cell
        for index, value in enumerate(grid.data().elements[:, upper]):
            if value == vertex1:
                local_vertex1 = index
                break
        else:
            local_vertex1 = -1

        for index, value in enumerate(grid.data().elements[:, lower]):
            if value == vertex2:
                local_vertex2 = index
                break
        else:
            local_vertex2 = -1

        # Numbering of barycentric cells on reference edge:
        # 6 * coarse_element_number + 2 * coarse_vertex_number
        bary_upper_minus = 6 * upper + 2 * local_vertex1
        bary_upper_plus = 6 * upper + 2 * local_vertex1 + 1
        bary_lower_minus = 6 * lower + 2 * local_vertex2
        bary_lower_plus = 6 * lower + 2 * local_vertex2 + 1

        # Starting from the upper minus cell or the lower minus cell, append in
        # anti clock-wise order the barycentric cells that follor
        vertex_edges1, sorted_edges1, number_of_cells1, ref_edge1 = _get_barycentric_edges_associated_to_vertex(
            vertex1, bary_vertex_to_edge, bary_upper_minus, bary_grid
        )
        vertex_edges2, sorted_edges2, number_of_cells2, ref_edge2 = _get_barycentric_edges_associated_to_vertex(
            vertex2, bary_vertex_to_edge, bary_lower_minus, bary_grid
        )

        aux_values, aux_bary_dofs, aux_coarse_dofs = _get_bary_coefficients(
            edge_lengths,
            vertex_edges1,
            vertex_edges2,
            sorted_edges1,
            sorted_edges2,
            bary_grid,
            local2global,
            number_of_cells1,
            number_of_cells2,
            global_dof_index,
            ref_edge1,
            ref_edge2,
        )

        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs

        aux_values, aux_bary_dofs, aux_coarse_dofs = _get_coefficients_reference_edge(
            edge_lengths,
            bary_grid,
            local2global,
            global_dof_index,
            bary_upper_minus,
            bary_upper_plus,
            bary_lower_minus,
            bary_lower_plus,
        )

        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs

    nentries = len(coarse_dofs)
    np_coarse_dofs = _np.zeros(nentries, dtype=_np.uint32)
    np_bary_dofs = _np.zeros(nentries, dtype=_np.uint32)
    np_values = _np.zeros(nentries, dtype=_np.float64)

    np_coarse_dofs[:] = coarse_dofs
    np_bary_dofs[:] = bary_dofs
    np_values[:] = values

    dof_transformation = coo_matrix(
        (np_values, (np_bary_dofs, np_coarse_dofs)),
        shape=(3 * bary_support_size, coarse_space.global_dof_count),
        dtype=_np.float64,
    ).tocsr()

    return (
        dof_transformation,
        support,
        normal_multipliers,
        local2global,
        local_multipliers,
    )


@_numba.njit(cache=True)
def _compute_rwg0_space_data(
    support,
    edge_neighbors,
    edge_neighbors_ptr,
    element_edges,
    number_of_elements,
    number_of_edges,
    include_boundary_dofs,
    truncate_at_segment_edge,
):
    """Compute the local2global map for the space."""
    local2global_map = _np.zeros((number_of_elements, 3), dtype=_np.uint32)
    local_multipliers = _np.zeros((number_of_elements, 3), dtype=_np.float64)
    edge_dofs = -_np.ones(number_of_edges, dtype=_np.int32)
    dof_count = 0
    for element in _np.flatnonzero(support):
        has_dof = False
        for local_index in range(3):
            edge_index = element_edges[local_index, element]
            if edge_dofs[edge_index] != -1:
                has_dof = True
            else:
                current_neighbors = edge_neighbors[edge_neighbors_ptr[edge_index] : edge_neighbors_ptr[1 + edge_index]]
                supported_neighbors = [e for e in current_neighbors if support[e]]
                if len(supported_neighbors) == 2:
                    if edge_dofs[edge_index]:
                        edge_dofs[edge_index] = dof_count
                        dof_count += 1
                    has_dof = True
                if len(supported_neighbors) == 1 and include_boundary_dofs:
                    if edge_dofs[edge_index]:
                        edge_dofs[edge_index] = dof_count
                        dof_count += 1
                    has_dof = True
                    if not truncate_at_segment_edge:
                        for cell in current_neighbors:
                            # Add the element to the support
                            support[cell] = True
        if not has_dof:
            # If the element has no DOFs, remove it from support
            support[element] = False

    for element_index in _np.flatnonzero(support):
        dofmap = -_np.ones(3, dtype=_np.int32)
        for local_index in range(3):
            edge_index = element_edges[local_index, element_index]
            if edge_dofs[edge_index] != -1:
                dofmap[local_index] = edge_dofs[edge_index]

                current_neighbors = edge_neighbors[edge_neighbors_ptr[edge_index] : edge_neighbors_ptr[1 + edge_index]]
                supported_neighbors = [e for e in current_neighbors if support[e]]

                if len(supported_neighbors) == 1:
                    local_multipliers[element_index, local_index] = 1
                else:
                    # Assign 1 or -1 depending on element index
                    local_multipliers[element_index, local_index] = (
                        1 if element_index == min(supported_neighbors) else -1
                    )

        # For every zero local multiplier assign an existing global dof
        # in this element. This does not change the result as zero multipliers
        # do not contribute. But it allows us not to have to distinguish between
        # existing and non existing dofs later on.
        first_nonzero = 0
        for local_index in range(3):
            if local_multipliers[element_index, local_index] != 0:
                first_nonzero = local_index
                break

        for local_index in range(3):
            if local_multipliers[element_index, local_index] == 0:
                dofmap[local_index] = dofmap[first_nonzero]
        local2global_map[element_index, :] = dofmap

    return dof_count, support, local2global_map, local_multipliers


@_numba.njit(cache=True)
def generate_rwg0_map(grid_data, support_elements, local_coords, coeffs):
    """Actually generate the sparse matrix data."""
    number_of_elements = len(support_elements)

    coarse_dofs = _np.empty(3 * 18 * number_of_elements, dtype=_np.uint32)
    bary_dofs = _np.empty(3 * 18 * number_of_elements, dtype=_np.uint32)
    values = _np.empty(3 * 18 * number_of_elements, dtype=_np.float64)

    # Iterate through the global dofs and fill up the
    # corresponding coefficients.

    count = 0

    for index, elem_index in enumerate(support_elements):
        # Compute all the local vertices
        local_vertices = grid_data.local2global(elem_index, local_coords)
        l1 = _np.linalg.norm(local_vertices[:, 6] - local_vertices[:, 4])
        l2 = _np.linalg.norm(local_vertices[:, 6] - local_vertices[:, 3])
        l3 = _np.linalg.norm(local_vertices[:, 6] - local_vertices[:, 5])
        l4 = _np.linalg.norm(local_vertices[:, 6] - local_vertices[:, 2])
        l5 = _np.linalg.norm(local_vertices[:, 6] - local_vertices[:, 1])
        l6 = _np.linalg.norm(local_vertices[:, 6] - local_vertices[:, 0])
        le1 = _np.linalg.norm(local_vertices[:, 2] - local_vertices[:, 0])
        le2 = _np.linalg.norm(local_vertices[:, 4] - local_vertices[:, 0])
        le3 = _np.linalg.norm(local_vertices[:, 4] - local_vertices[:, 2])

        outer_edges = [le1, le2, le3]

        dof_mult = _np.array(
            [
                [le1, l6, l5],
                [l4, le1, l5],
                [le3, l4, l2],
                [l1, le3, l2],
                [le2, l1, l3],
                [l6, le2, l3],
            ]
        )

        # Assign the dofs for the six barycentric elements

        bary_elements = _np.arange(6) + 6 * index
        for local_dof in range(3):
            coarse_dof = 3 * index + local_dof
            bary_coeffs = coeffs[local_dof]
            dof_coeffs = bary_coeffs * outer_edges[local_dof] / dof_mult
            coarse_dofs[count : count + 18] = coarse_dof
            bary_dofs[count : count + 18] = _np.arange(3 * bary_elements[0], 3 * bary_elements[0] + 18)
            values[count : count + 18] = dof_coeffs.ravel()
            count += 18
    return coarse_dofs, bary_dofs, values


@_numba.njit()
def _numba_rwg0_evaluate(
    element_index,
    shapeset_evaluate,
    local_coordinates,
    grid_data,
    local_multipliers,
    normal_multipliers,
):
    """Evaluate the basis on an element."""
    reference_values = shapeset_evaluate(local_coordinates)
    npoints = local_coordinates.shape[1]
    result = _np.empty((3, 3, npoints), dtype=_np.float64)

    edge_lengths = _np.empty(3, dtype=_np.float64)
    edge_lengths[0] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[0, element_index]]
        - grid_data.vertices[:, grid_data.elements[1, element_index]]
    )
    edge_lengths[1] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[2, element_index]]
        - grid_data.vertices[:, grid_data.elements[0, element_index]]
    )
    edge_lengths[2] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[1, element_index]]
        - grid_data.vertices[:, grid_data.elements[2, element_index]]
    )

    for index in range(3):
        result[:, index, :] = (
            local_multipliers[element_index, index]
            * edge_lengths[index]
            / grid_data.integration_elements[element_index]
            * grid_data.jacobians[element_index].dot(reference_values[:, index, :])
        )
    return result


@_numba.njit()
def _numba_snc0_evaluate(
    element_index,
    shapeset_evaluate,
    local_coordinates,
    grid_data,
    local_multipliers,
    normal_multipliers,
):
    """Evaluate the basis on an element."""
    reference_values = shapeset_evaluate(local_coordinates)
    npoints = local_coordinates.shape[1]
    result = _np.empty((3, 3, npoints), dtype=_np.float64)
    tmp = _np.empty((3, 3, npoints), dtype=_np.float64)
    normal = grid_data.normals[element_index] * normal_multipliers[element_index]

    edge_lengths = _np.empty(3, dtype=_np.float64)
    edge_lengths[0] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[0, element_index]]
        - grid_data.vertices[:, grid_data.elements[1, element_index]]
    )
    edge_lengths[1] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[2, element_index]]
        - grid_data.vertices[:, grid_data.elements[0, element_index]]
    )
    edge_lengths[2] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[1, element_index]]
        - grid_data.vertices[:, grid_data.elements[2, element_index]]
    )

    for index in range(3):
        tmp[:, index, :] = (
            local_multipliers[element_index, index]
            * edge_lengths[index]
            / grid_data.integration_elements[element_index]
            * grid_data.jacobians[element_index].dot(reference_values[:, index, :])
        )

    result[0, :, :] = normal[1] * tmp[2, :, :] - normal[2] * tmp[1, :, :]
    result[1, :, :] = normal[2] * tmp[0, :, :] - normal[0] * tmp[2, :, :]
    result[2, :, :] = normal[0] * tmp[1, :, :] - normal[1] * tmp[0, :, :]

    return result


@_numba.njit
def _numba_snc0_surface_curl(
    element_index,
    shapeset_gradient,
    local_coordinates,
    grid_data,
    local_multipliers,
    normal_multipliers,
):
    """Evaluate the curl on an element."""
    normal = grid_data.normals[element_index] * normal_multipliers[element_index]
    reference_derivatives = shapeset_gradient(local_coordinates)
    jac_inv_t = grid_data.jac_inv_trans[element_index]
    derivatives = jac_inv_t @ reference_derivatives @ jac_inv_t.T
    reference_values = (
        normal[0] * (derivatives[2, 1] - derivatives[1, 2])
        + normal[1] * (derivatives[0, 2] - derivatives[2, 0])
        + normal[2] * (derivatives[1, 0] - derivatives[0, 1])
    )

    result = _np.empty(3, dtype=_np.float64)

    edge_lengths = _np.empty(3, dtype=_np.float64)
    edge_lengths[0] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[0, element_index]]
        - grid_data.vertices[:, grid_data.elements[1, element_index]]
    )
    edge_lengths[1] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[2, element_index]]
        - grid_data.vertices[:, grid_data.elements[0, element_index]]
    )
    edge_lengths[2] = _np.linalg.norm(
        grid_data.vertices[:, grid_data.elements[1, element_index]]
        - grid_data.vertices[:, grid_data.elements[2, element_index]]
    )

    for index in range(3):
        result[index] = local_multipliers[element_index, index] * edge_lengths[index] * reference_values
    return result



"""Define scalar spaces on the dual grid."""

import numpy as _np
from bempp_cl.api import log

from .scalar_spaces import (
    p1_continuous_function_space,
    p0_discontinuous_function_space,
    _numba_p0_surface_gradient,
    _numba_p1_surface_gradient,
)


def dual0_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=False,
    truncate_at_segment_edge=False,
):
    """Define a space of DP0 functions on the dual grid."""
    from .space import SpaceBuilder, invert_local2global
    from scipy.sparse import coo_matrix

    coarse_space = p1_continuous_function_space(
        grid,
        support_elements,
        segments,
        swapped_normals,
        include_boundary_dofs=include_boundary_dofs,
        truncate_at_segment_edge=truncate_at_segment_edge,
    )

    number_of_support_elements = coarse_space.number_of_support_elements

    bary_support_elements = 6 * _np.repeat(coarse_space.support_elements, 6) + _np.tile(
        _np.arange(6), number_of_support_elements
    )

    bary_grid = grid.barycentric_refinement
    bary_support_size = len(bary_support_elements)

    support = _np.zeros(bary_grid.number_of_elements, dtype=_np.bool_)
    support[bary_support_elements] = True

    local2global = _np.zeros((bary_grid.number_of_elements, 1), dtype="uint32")
    local_multipliers = _np.zeros((bary_grid.number_of_elements, 1), dtype="uint32")

    local2global[support] = _np.arange(bary_support_size).reshape(bary_support_size, 1)

    local_multipliers[support] = 1
    global2local = invert_local2global(local2global, local_multipliers)

    _bary_dofs = []
    _coarse_dofs = []

    support_numbers = {j: i for i, j in enumerate(coarse_space.support_elements)}

    for global_dof_index in range(coarse_space.global_dof_count):
        local_dofs = coarse_space.global2local[global_dof_index]

        for face, vertex in local_dofs:
            if support[face] or not truncate_at_segment_edge:
                face_n = support_numbers[face]
                _coarse_dofs.append(global_dof_index)
                _coarse_dofs.append(global_dof_index)
                _bary_dofs.append(6 * face_n + (2 * vertex - 1) % 6)
                _bary_dofs.append(6 * face_n + 2 * vertex)

    nentries = len(_bary_dofs)

    coarse_dofs = _np.zeros(nentries, dtype=_np.uint32)
    coarse_dofs[:] = _coarse_dofs

    bary_dofs = _np.zeros(nentries, dtype=_np.uint32)
    bary_dofs[:] = _bary_dofs

    values = _np.ones(nentries, dtype=_np.float64)

    dof_transformation = coo_matrix(
        (values, (bary_dofs, coarse_dofs)),
        shape=(bary_support_size, coarse_space.global_dof_count),
        dtype=_np.float64,
    ).tocsr()

    collocation_points = _np.array([[1.0 / 3], [1.0 / 3]])

    return (
        SpaceBuilder(bary_grid)
        .set_codomain_dimension(1)
        .set_support(support)
        .set_order(0)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("p0_discontinuous")
        .set_identifier("p0_discontinuous")
        .set_local2global(local2global)
        .set_global2local(global2local)
        .set_local_multipliers(local_multipliers)
        .set_collocation_points(collocation_points)
        .set_dof_transformation(dof_transformation)
        .set_numba_surface_gradient(_numba_p0_surface_gradient)
        .build()
    )


def dual1_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=None,
    truncate_at_segment_edge=False,
):
    """Define a space of DP1 functions on the dual grid."""
    from .space import SpaceBuilder, invert_local2global
    from scipy.sparse import coo_matrix

    if include_boundary_dofs is not None:
        log("Setting include_boundary_dofs has no effect on this space type.", "warning")

    coarse_space = p0_discontinuous_function_space(grid, support_elements, segments, swapped_normals)

    coarse_support = _np.zeros(grid.number_of_elements, dtype=_np.bool_)
    coarse_support[coarse_space.support_elements] = True

    nentries = 0
    for global_dof_index in range(coarse_space.global_dof_count):
        local_dofs = coarse_space.global2local[global_dof_index]
        element_index = local_dofs[0][0]

        # 1 at barycentre of the triangle
        nentries += 6
        # 1/2 at the centre of each edge
        for e in range(3):
            edge = coarse_space.grid.element_edges[e][element_index]
            for neighbour in coarse_space.grid.edge_neighbors[edge]:
                if not truncate_at_segment_edge:
                    coarse_support[neighbour] = True
                if coarse_support[neighbour]:
                    nentries += 2
        # 1/num_coarse_triangles_at_vertex at each vertex
        for v in range(3):
            vertex = coarse_space.grid.elements[v][element_index]
            start = coarse_space.grid.vertex_neighbors.indexptr[vertex]
            end = coarse_space.grid.vertex_neighbors.indexptr[vertex + 1]
            neighbours = coarse_space.grid.vertex_neighbors.indices[start:end]
            for neighbour in neighbours:
                if not truncate_at_segment_edge:
                    coarse_support[neighbour] = True
                if coarse_support[neighbour]:
                    nentries += 2

    support_elements = [i for i, j in enumerate(coarse_support) if j]
    support_numbers = {j: i for i, j in enumerate(support_elements)}

    coarse_dofs = _np.zeros(nentries, dtype=_np.uint32)
    bary_dofs = _np.zeros(nentries, dtype=_np.uint32)
    values = _np.zeros(nentries, dtype=_np.float64)

    count = 0
    for global_dof_index in range(coarse_space.global_dof_count):
        local_dofs = coarse_space.global2local[global_dof_index]
        element_index = local_dofs[0][0]

        # 1 at barycentre of the triangle
        if coarse_support[element_index]:
            face_n = support_numbers[element_index]
            for n in [1, 5, 7, 11, 13, 17]:
                bary_dofs[count] = 6 * 3 * face_n + n
                coarse_dofs[count] = global_dof_index
                values[count] = 1
                count += 1
        # 1/2 at the centre of each edge
        for e in range(3):
            edge = coarse_space.grid.element_edges[e][element_index]
            for neighbour in coarse_space.grid.edge_neighbors[edge]:
                if coarse_support[neighbour]:
                    face_n = support_numbers[neighbour]
                    for i, dofs in enumerate([[1, 5], [13, 17], [7, 11]]):
                        if coarse_space.grid.element_edges[i][neighbour] == edge:
                            for n in dofs:
                                bary_dofs[count] = 6 * 3 * face_n + n
                                coarse_dofs[count] = global_dof_index
                                values[count] = 0.5
                                count += 1
                            break
        # 1/num_coarse_triangles_at_vertex at each vertex
        for v in range(3):
            vertex = coarse_space.grid.elements[v][element_index]
            start = coarse_space.grid.vertex_neighbors.indexptr[vertex]
            end = coarse_space.grid.vertex_neighbors.indexptr[vertex + 1]
            neighbour_count = end - start
            neighbours = coarse_space.grid.vertex_neighbors.indices[start:end]
            for neighbour in neighbours:
                if coarse_support[neighbour]:
                    face_n = support_numbers[neighbour]
                    for i, dofs in enumerate([[0, 15], [3, 6], [9, 12]]):
                        if coarse_space.grid.elements[i][neighbour] == vertex:
                            for n in dofs:
                                bary_dofs[count] = 6 * 3 * face_n + n
                                coarse_dofs[count] = global_dof_index
                                values[count] = 1 / neighbour_count
                                count += 1
                            break

    bary_grid = grid.barycentric_refinement

    number_of_support_elements = len(support_elements)

    coarse_elements = _np.array(support_elements, dtype=_np.uint32)

    bary_support_elements = 6 * _np.repeat(coarse_elements, 6) + _np.tile(_np.arange(6), number_of_support_elements)

    bary_support_size = len(bary_support_elements)

    support = _np.zeros(bary_grid.number_of_elements, dtype=_np.bool_)
    support[bary_support_elements] = True

    local2global = _np.zeros((bary_grid.number_of_elements, 3), dtype="uint32")
    local_multipliers = _np.zeros((bary_grid.number_of_elements, 3), dtype="uint32")

    local2global[support] = _np.arange(3 * bary_support_size).reshape(bary_support_size, 3)

    local_multipliers[support] = 1
    global2local = invert_local2global(local2global, local_multipliers)

    dof_transformation = coo_matrix(
        (values, (bary_dofs, coarse_dofs)),
        shape=(3 * bary_support_size, coarse_space.global_dof_count),
        dtype=_np.float64,
    ).tocsr()

    return (
        SpaceBuilder(bary_grid)
        .set_codomain_dimension(1)
        .set_support(support)
        .set_order(1)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("p1_discontinuous")
        .set_identifier("p1_discontinuous")
        .set_local2global(local2global)
        .set_global2local(global2local)
        .set_local_multipliers(local_multipliers)
        .set_dof_transformation(dof_transformation)
        .set_numba_surface_gradient(_numba_p1_surface_gradient)
        .build()
    )



"""Definition of shapesets on the reference element."""

import numba as _numba
import numpy as _np


class Shapeset(object):
    """Definintion of a shapeset on a reference element."""

    def __init__(self, identifier):
        """Define a shapeset from its identifier."""
        data = _SHAPESETS[identifier]

        self._evaluate = data["evaluate"]
        self._gradient = data["gradient"]
        self._number_of_shape_functions = data["number_of_shape_functions"]
        self._identifier = data["identifier"]
        self._dimension = data["dimension"]

    @property
    def evaluate(self):
        """Evaluate the shapeset."""
        return self._evaluate

    @property
    def gradient(self):
        """Evaluate the gradient of the shapeset."""
        return self._gradient

    @property
    def number_of_shape_functions(self):
        """Return the number of shape functions."""
        return self._number_of_shape_functions

    @property
    def identifier(self):
        """Return identifier of this shapeset."""
        return self._identifier

    @property
    def dimension(self):
        """Return the dimension of the shapeset."""
        return self._dimension


@_numba.njit
def _p0_shapeset_evaluate(local_coordinates):
    """Evaluate P0 shapeset."""
    return _np.ones((1, 1, local_coordinates.shape[1]), dtype=local_coordinates.dtype)


@_numba.njit
def _p0_shapeset_gradient(local_coordinates):
    """Evaluate P0 gradient."""
    return _np.zeros((1, 2, 1, local_coordinates.shape[1]), dtype=local_coordinates.dtype)


@_numba.njit
def _p1_disc_shapeset_evaluate(local_coordinates):
    """Evaluate P1 discontinuous shapeset."""
    return _np.expand_dims(
        _np.vstack(
            (
                1 - local_coordinates[0, :] - local_coordinates[1, :],
                local_coordinates[0, :],
                local_coordinates[1, :],
            )
        ),
        0,
    )


@_numba.njit
def _p1_disc_shapeset_gradient(local_coordinates):
    """Evaluate P1 discontinuous shapeset gradient."""
    dtype = local_coordinates.dtype
    grad = _np.zeros((1, 2, 3, local_coordinates.shape[1]), dtype=dtype)
    grad[0, 0, :, :] = _np.vstack(
        (
            -_np.ones(local_coordinates.shape[1], dtype=dtype),
            _np.ones(local_coordinates.shape[1], dtype=dtype),
            _np.zeros(local_coordinates.shape[1], dtype=dtype),
        )
    )
    grad[0, 1, :, :] = _np.vstack(
        (
            -_np.ones(local_coordinates.shape[1], dtype=dtype),
            _np.zeros(local_coordinates.shape[1], dtype=dtype),
            _np.ones(local_coordinates.shape[1], dtype=dtype),
        )
    )
    return grad


@_numba.njit
def _rwg0_shapeset_evaluate(local_coordinates):
    """Evaluate RWG 0 shapeset."""
    dtype = local_coordinates.dtype
    vals = _np.zeros((2, 3, local_coordinates.shape[1]), dtype=dtype)
    vals[0, :, :] = _np.vstack((local_coordinates[0], local_coordinates[0] - 1, local_coordinates[0]))
    vals[1, :, :] = _np.vstack((local_coordinates[1] - 1, local_coordinates[1], local_coordinates[1]))
    return vals


@_numba.njit
def _rwg0_shapeset_gradient(local_coordinates):
    """Evaluate RWG 0 shapeset gradient."""
    dtype = local_coordinates.dtype
    npoints = local_coordinates.shape[1]
    grad = _np.zeros((2, 2, 3, npoints), dtype=_np.dtype)
    # Mixed derivates (derivative in direction j of
    # component i with i neq j) are zero.
    grad[0, 0, :, :] = _np.ones((3, npoints), dtype=dtype)
    grad[1, 1, :, :] = _np.ones((3, npoints), dtype=dtype)
    return grad


@_numba.njit
def _snc0_shapeset_gradient(local_coordinates):
    """Evaluate SNC 0 shapeset gradient."""
    dtype = local_coordinates.dtype
    temp = _np.eye(2, dtype=dtype)
    grad = _np.zeros((2, 2), dtype=dtype)
    grad[0] = -temp[1]
    grad[1] = temp[0]
    return grad


_SHAPESETS = {
    "p0_discontinuous": {
        "evaluate": _p0_shapeset_evaluate,
        "gradient": _p0_shapeset_gradient,
        "number_of_shape_functions": 1,
        "identifier": "p0_discontinuous",
        "dimension": 1,
    },
    "p1_discontinuous": {
        "evaluate": _p1_disc_shapeset_evaluate,
        "gradient": _p1_disc_shapeset_gradient,
        "number_of_shape_functions": 3,
        "identifier": "p1_discontinuous",
        "dimension": 1,
    },
    "rwg0": {
        "evaluate": _rwg0_shapeset_evaluate,
        "gradient": _rwg0_shapeset_gradient,
        "number_of_shape_functions": 3,
        "identifier": "rwg0",
        "dimension": 2,
    },
    "snc0": {
        "evaluate": _rwg0_shapeset_evaluate,
        "gradient": _snc0_shapeset_gradient,
        "number_of_shape_functions": 3,
        "identifier": "snc0",
        "dimension": 2,
    },
}



import numpy as np
from electroacPy.general import freqop, geometry, io
import matplotlib.pyplot as plt


class laserVibrometry_UFF:
    def __init__(self, file_path, spatial_rotation, freq_array, 
                 useAverage=False, inputVoltage=1):
        
        """
        Import UFF file to load acceleration data.
        """
        
        self.file_path = file_path
        self.spatial_rotation = spatial_rotation
        self.identifier = "PLV"

        # load and decimate data to match frequency axis
        freq_meas, acc_meas, v_meas, point_cloud = io.loadUFF(file_path)
        freq_dec, index_dec = freqop.decimate_frequency_axis(freq_meas, freq_array)
        acc_dec = acc_meas[:, index_dec] * inputVoltage
        v_dec = v_meas[:, index_dec] * inputVoltage

        self.acc_dec = acc_dec
        self.freq_dec = freq_dec
        
        # useful data
        if np.all(v_dec == 0) is True or np.array(v_dec).size == 0:
            self.v_point = acc_dec / freqop.laplace(freq_dec)       # velocity at each points (nPoints, Nfft)
        else:
            self.v_point = acc_dec / freqop.laplace(freq_dec) # v_dec
        self.v           = np.ones(len(freq_array), dtype=complex)  # decoy for post processing ?
        self.point_cloud = geometry.rotatePointCloud(geometry.recenterZero(point_cloud), spatial_rotation)

        self.v_mean = np.zeros([len(freq_dec)], dtype=complex)
        for f in range(len(freq_dec)):
            self.v_mean[f] = np.mean(self.v_point[:, f])

        if useAverage is True:
            for f in range(len(freq_dec)):
                self.v_point[:, f] = self.v_mean[f]

        # identifiers
        self.ref2bem   = None
        self.poly_data = True  # is class from polytech?

    def plot_pointCloud(self):
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
        ax.scatter(self.point_cloud[1:, 0]*1e2, self.point_cloud[1:, 1]*1e2, self.point_cloud[1:, 2]*1e2)
        ax.scatter(self.point_cloud[0, 0]*1e2, self.point_cloud[0, 1]*1e2, self.point_cloud[0, 2]*1e2, 'r')
        ax.axis('equal')
        ax.set(xlabel='x [cm]', ylabel='y [cm]', zlabel='z [cm]', title='Vibrometry points')
        return plt.show()

    def plot_velocity(self, point_index):
        fig, ax = plt.subplots(2, 1)
        ax[0].semilogx(self.freq_dec, np.abs(self.v_point[point_index, :]))
        ax[1].semilogx(self.freq_dec, np.angle(self.v_point[point_index, :]))
        for i in range(2):
            ax[i].set(xlabel='Frequency [Hz]', ylabel='Velocity [m/s]')
            ax[i].grid(which='both')
        plt.tight_layout()
        return plt.show()

    def plotXVA(self):
        """
        Plot the displacement, velocity, and acceleration frequency responses.

        Returns
        -------
        None

        """
        s = freqop.laplace(self.freq_dec)
        Hx = self.v_mean / s
        Hv = self.v_mean
        Ha = self.v_mean * s

        fig, ax = plt.subplots(3, 1)
        ax[0].semilogx(self.freq_dec, np.abs(Hx * 1e3), label='Displacement')
        ax[1].semilogx(self.freq_dec, np.abs(Hv), label='Velocity')
        ax[2].semilogx(self.freq_dec, np.abs(Ha), label='Acceleration')
        ax[2].set(xlabel="Frequency [Hz]")
        ax[0].set(ylabel="mm")
        ax[1].set(ylabel="m/s")
        ax[2].set(ylabel="m/s^2")
        for i in range(3):
            ax[i].grid(which='both')
            ax[i].legend(loc='best')
        plt.tight_layout()
        return plt.show()



class laserVibrometry:
    def __init__(self, Hv, X, inputVoltage=1):
        """
        
        Import user-defined vib data in numpy arrays.        
        
        Parameters
        ----------
        Hv : numpy array
            Velocity data (nPoints, NFFT). Nfft should corresponds to the number of points
            in your simulation.
        X : numpy array
            Position data (nPoints, 3) in cartesian coordinates.
    
        Returns
        -------
        None.

        """
        self.identifier = "PLV"
        self.v_point = Hv * inputVoltage 
        self.point_cloud = X
        
        self.v = np.ones(Hv.shape[1], dtype=complex) 
        self.v_mean = np.mean(self.v_point, 1)

        # identifiers
        self.ref2bem   = None
        self.poly_data = True  # is class from polytech?

    def plot_pointCloud(self):
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
        ax.scatter(self.point_cloud[1:, 0]*1e2, self.point_cloud[1:, 1]*1e2, self.point_cloud[1:, 2]*1e2)
        ax.scatter(self.point_cloud[0, 0]*1e2, self.point_cloud[0, 1]*1e2, self.point_cloud[0, 2]*1e2, 'r')
        ax.axis('equal')
        ax.set(xlabel='x [cm]', ylabel='y [cm]', zlabel='z [cm]', title='Vibrometry points')
        return plt.show()

    def plot_velocity(self, point_index):
        fig, ax = plt.subplots(2, 1)
        ax[0].semilogx(self.freq_dec, np.abs(self.v_point[point_index, :]))
        ax[1].semilogx(self.freq_dec, np.angle(self.v_point[point_index, :]))
        for i in range(2):
            ax[i].set(xlabel='Frequency [Hz]', ylabel='Velocity [m/s]')
            ax[i].grid(which='both')
        plt.tight_layout()
        return plt.show()


# pylint: disable-msg=too-many-arguments
"""Definition of Grid functions in Bempp.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/grid_functions.html
"""

import numba as _numba
import numpy as _np


def callable(*args, complex=False, jit=True, parameterized=False, vectorized=False):
    """Wrap a callable for Bempp."""
    if vectorized:

        def wrap(f):
            """Wrap a functions (vectorised)."""
            if parameterized:

                def wrapper_callable(x, n, domain_index, res, parameters):
                    """Call the wrapped callable."""
                    f(x, n, domain_index, res, parameters)

            else:

                def wrapper_callable(x, n, domain_index, res, parameters):
                    """Call the wrapped callable."""
                    f(x, n, domain_index, res)

            wrapper_callable.bempp_type = "complex" if complex else "real"
            wrapper_callable.bempp_vectorized = True

            return wrapper_callable

    else:
        if complex:
            wrap_type = _numba.complex128[:]
        else:
            wrap_type = _numba.float64[:]

        signature_parameterized = _numba.void(_numba.float64[:], _numba.float64[:], _numba.uint32, wrap_type, wrap_type)
        signature_not_parameterized = _numba.void(_numba.float64[:], _numba.float64[:], _numba.uint32, wrap_type)

        if parameterized:
            signature = signature_parameterized
        else:
            signature = signature_not_parameterized

        def wrap(f):
            """Wrap a function."""
            if not jit:
                if parameterized:

                    def wrapper_callable(x, n, domain_index, res, parameters):
                        """Call using object mode."""
                        with _numba.objmode():
                            f(x, n, domain_index, res, parameters)

                else:

                    def wrapper_callable(x, n, domain_index, res, parameters):
                        """Call using object mode."""
                        with _numba.objmode():
                            f(x, n, domain_index, res)

            else:
                f_jit = _numba.njit(signature)(f)
                if parameterized:

                    def wrapper_callable(x, n, domain_index, res, parameters):
                        """Call a wrapped function."""
                        f_jit(x, n, domain_index, res, parameters)

                else:

                    def wrapper_callable(x, n, domain_index, res, parameters):
                        """Call a wrapped function."""
                        f_jit(x, n, domain_index, res)

            njit_wrapper = _numba.njit(signature_parameterized)(wrapper_callable)

            njit_wrapper.bempp_type = "complex" if complex else "real"
            njit_wrapper.bempp_vectorized = False

            return njit_wrapper

    if not args:
        return wrap
    else:
        return wrap(args[0])


def real_callable(*args, jit=True):
    """Wrap function as a real Numba callable."""
    return callable(*args, complex=False, jit=jit)


def complex_callable(*args, jit=True):
    """Wrap function as a complex Numba callable."""
    return callable(*args, complex=True, jit=jit)


class GridFunction(object):
    """
    Representation of functions on a grid.

    Attributes
    ----------
    coefficients : np.ndarray
        Return or set the vector of coefficients.
    component_count : int
        Return the number of components of the grid
        function values.
    space : bemp.api.space.Space
        Return the space over which the GridFunction is defined.
    grid : bempp_cl.api.grid.Grid
        Return the underlying grid.
    parameters : bempp_cl.api.ParameterList
        Return the set of parameters.
    representation : string
        Return 'primal' if the coefficients of the Gridfunction
        are known. Return 'dual' if only the coefficients in the
        dual space are known.

    """

    def __init__(
        self,
        space,
        dual_space=None,
        fun=None,
        coefficients=None,
        projections=None,
        parameters=None,
        function_parameters=None,
    ):
        """
        Construct a grid function.

        A grid function can be initialized in three different ways.

        1. By providing a Python callable. Any Python callable of the
           following form is valid.::

                callable(x,n,domain_index,result)

           Here, x, n, and result are all numpy arrays. x contains the current
           evaluation point, n the associated outward normal direction and
           result is a numpy array that will store the result of the Python
           callable. The variable domain_index stores the index of the
           subdomain on which x lies (default 0). This makes it possible to
           define different functions for different subdomains.

           The following example defines input data that is the inner product
           of the coordinate x with the normal direction n.::

                fun(x,n,domain_index,result):
                    result[0] =  np.dot(x,n)

        2. By providing a vector of coefficients at the nodes. This is
           preferable if the coefficients of the data are coming from an
           external code.

        3. By providing a vector of projection data and a corresponding
           dual space.

        Parameters
        ----------
        space : bempp_cl.api.space.Space
            The space over which the GridFunction is defined.
        dual_space : bempp_cl.api.Space
            A representation of the dual space. If not specified
            then space == dual_space is assumed (optional).
        fun : callable
            A Python function from which the GridFunction is constructed
            (optional).
        coefficients : np.ndarray
            A 1-dimensional array with the coefficients of the GridFunction
            at the interpolatoin points of the space (optional).
        projections : np.ndarray
            A 1-dimensional array with the projections of the GridFunction
            onto a dual space (optional).
        parameters : bempp_cl.api.ParameterList
            A ParameterList object used for the assembly of
            the GridFunction (optional).
        function_parameters : np.ndarray
            Numpy array containing optional parameters for the callable
            to discretise.

        Notes
        -----
        * Only one of projections, coefficients, or fun is allowed as
          parameter.

        Examples
        --------
        To create a GridFunction from a Python callable my_fun use

        >>> grid_function = GridFunction(space, fun=my_fun)

        To create a GridFunction from a vector of coefficients coeffs use

        >>> grid_function = GridFunction(space,coefficients=coeffs)

        To create a GridFunction from a vector of projections proj use

        >>> grid_function = GridFunction(
                space,dual_space=dual_space, projections=proj)

        """
        from bempp_cl.api.utils.helpers import assign_parameters
        from bempp_cl.api.space.space import return_compatible_representation

        self._space = None
        self._dual_space = None
        self._coefficients = None
        self._grid_coefficients = None
        self._projections = None
        self._representation = None

        if dual_space is None:
            dual_space = space

        self._space, self._dual_space = space, dual_space

        # Now check that space and dual are defined over same grid
        # with the same normal directions. If one space is barycentric,
        # need to take this into account.

        comp_domain, comp_dual = return_compatible_representation(space, dual_space)
        self._comp_domain = comp_domain
        self._comp_dual = comp_dual

        if not comp_domain.grid == comp_dual.grid or not _np.all(
            comp_domain.normal_multipliers == comp_dual.normal_multipliers
        ):
            raise ValueError("Space and dual space must be defined on the " + "same grid with same normal directions.")

        self._parameters = assign_parameters(parameters)

        if sum(1 for e in [fun, coefficients, projections] if e is not None) != 1:
            raise ValueError("Exactly one of 'fun', 'coefficients' or 'projections' " + "must be nonzero.")

        if coefficients is not None:
            self._coefficients = coefficients
            self._representation = "primal"

        if projections is not None:
            self._projections = projections
            self._representation = "dual"

        if fun is not None:
            from bempp_cl.api.integration.triangle_gauss import rule

            points, weights = rule(self._parameters.quadrature.regular)

            if function_parameters is None:
                function_parameters = _np.array([], dtype="float64")

            if fun.bempp_type == "real":
                dtype = "float64"
            else:
                dtype = "complex128"
                if function_parameters is not None:
                    function_parameters = function_parameters.astype("complex128")

            grid_projections = _np.zeros(comp_dual.grid_dof_count, dtype=dtype)
            if not fun.bempp_vectorized:
                # Callable is not vectorized

                # Create a Numba callable from the function
                _project_function(
                    fun,
                    comp_dual.grid.data("double"),
                    comp_dual.support_elements,
                    comp_dual.local2global,
                    comp_dual.local_multipliers,
                    comp_dual.normal_multipliers,
                    comp_dual.numba_evaluate,
                    comp_dual.shapeset.evaluate,
                    points,
                    weights,
                    comp_domain.codomain_dimension,
                    grid_projections,
                    function_parameters,
                )
            else:
                # Callable is vectorized

                (
                    global_quad_points,
                    global_normals,
                    global_domain_indices,
                ) = get_function_quadrature_information(
                    comp_dual.grid.data("double"),
                    comp_dual.support_elements,
                    comp_dual.normal_multipliers,
                    points,
                )
                function_data = _np.empty(
                    (comp_domain.codomain_dimension, global_quad_points.shape[1]),
                    dtype=dtype,
                )
                fun(
                    global_quad_points,
                    global_normals,
                    global_domain_indices,
                    function_data,
                    function_parameters,
                )
                _project_function_vectorized(
                    function_data,
                    comp_dual.grid.data("double"),
                    comp_dual.support_elements,
                    comp_dual.local2global,
                    comp_dual.local_multipliers,
                    comp_dual.normal_multipliers,
                    comp_dual.numba_evaluate,
                    comp_dual.shapeset.evaluate,
                    points,
                    weights,
                    comp_domain.codomain_dimension,
                    grid_projections,
                    function_parameters,
                )

            self._projections = comp_dual.dof_transformation.T @ grid_projections

            self._representation = "dual"

    @property
    def space(self):
        """Return space."""
        return self._space

    @property
    def dtype(self):
        """Return type."""
        if self._coefficients is not None:
            return self._coefficients.dtype
        if self._projections is not None:
            return self._projections.dtype

    @property
    def parameters(self):
        """Return parameters."""
        return self._parameters

    @property
    def dual_space(self):
        """Return dual space."""
        return self._dual_space

    @property
    def representation(self):
        """
        Return the representation of the grid function.

        If the grid function is given in terms of coefficients in the
        domain space return 'primal'. If the function is given by its
        projections return 'dual'.
        """
        return self._representation

    @property
    def coefficients(self):
        """Return coefficient vector."""
        if self._coefficients is None:
            from bempp_cl.api.utils.helpers import get_inverse_mass_matrix

            op = get_inverse_mass_matrix(self.space, self.dual_space)
            self._coefficients = op @ self._projections
            self._representation = "primal"

        return self._coefficients

    @property
    def grid_coefficients(self):
        """Return grid coefficients."""
        if self._grid_coefficients is None:
            self._grid_coefficients = self.space.dof_transformation @ self.coefficients

        return self._grid_coefficients

    @property
    def real(self):
        """Return a new grid function consisting of the real part of this function."""
        import numpy as np
        import bempp_cl.api

        if self.representation == "primal":
            return bempp_cl.api.GridFunction(
                space=self.space,
                dual_space=self.dual_space,
                coefficients=np.real(self.coefficients),
            )
        else:
            return bempp_cl.api.GridFunction(
                space=self.space,
                dual_space=self.dual_space,
                projections=np.real(self.projections()),
            )

    @property
    def imag(self):
        """Return a new grid function consisting of the imaginary part of this function."""
        import numpy as np
        import bempp_cl.api

        if self.representation == "primal":
            return bempp_cl.api.GridFunction(
                space=self.space,
                dual_space=self.dual_space,
                coefficients=np.imag(self.coefficients),
            )
        else:
            return bempp_cl.api.GridFunction(
                space=self.space,
                dual_space=self.dual_space,
                projections=np.imag(self.projections()),
            )

    @property
    def component_count(self):
        """Return number of components."""
        return self.space.codomain_dimension

    def projections(self, dual_space=None):
        """
        Compute the vector of projections onto the given dual space.

        Parameters
        ----------
        dual_space : bempp_cl.api.space.Space
            A representation of the dual space. If not specified
            then fun.dual_space is used.

        Returns
        -------
        out : np.ndarray
            A vector of projections onto the dual space.

        """
        from bempp_cl.api.utils.helpers import get_mass_matrix

        if dual_space is None:
            dual_space = self.dual_space

        if dual_space == self._dual_space and self._projections is not None:
            return self._projections

        ident = get_mass_matrix(self.space, dual_space)
        return ident * self.coefficients

    def project_to_space(self, space):
        """Return an L^2 projection on another space."""
        from bempp_cl.api.utils.helpers import get_mass_matrix

        ident = get_mass_matrix(self.space, space)

        return GridFunction(space, projections=ident @ self.coefficients)

    def plot(self, mode=None, transformation=None):
        """
        Plot the grid function.

        Attributes
        ----------
        mode : string
            One of 'element' or 'node'. If 'element' is chosen
            the color is determined by the mid-point of the faces
            of the grid. For 'vertices' the vertex values are
            chosen
        transformation : string or object
            One of 'real', 'imag', 'abs', 'log_abs' or
            'abs_squared' or a callable object.
            Describes the data transformation
            before plotting. For functions with vector values
            only 'abs', 'log_abs' or 'abs_squared' are allowed.
            If a callable object is given this is applied instead.
            It is important that the callable returns numpy arrays
            with the same number of dimensions as before.

        """
        from bempp_cl.api.external.viewers import visualize

        visualize(self, mode, transformation)

    def evaluate(self, element_index, local_coordinates):
        """Evaluate grid function on a single element."""
        # Get global dof ids and weights
        global_dofs = self.space.local2global[element_index]
        element_values = self.space.evaluate(element_index, local_coordinates)
        return _np.tensordot(element_values, self.grid_coefficients[global_dofs], axes=([1], [0]))

    def evaluate_on_element_centers(self):
        """Evaluate the grid function on all element centers."""
        local_coordinates = _np.array([[1.0 / 3], [1.0 / 3]])

        values = _np.zeros((self.component_count, self.space.grid.number_of_elements), dtype=self.dtype)

        for index in self.space.support_elements:
            local_values = self.evaluate(index, local_coordinates)
            values[:, index] = local_values.flat
        return values

    def evaluate_on_vertices(self):
        """
        Evaluate the grid function on all vertices.

        If a function is discontinuous across elements a weighted average
        of the element values at the vertices is taken.
        """
        grid = self.space.grid
        local_coordinates = _np.array([[0, 1, 0], [0, 0, 1]], dtype="float64")

        values = _np.zeros((self.component_count, grid.number_of_vertices), dtype=self.dtype)

        # Sum up the areas of all elements adjacent to the vertices
        vertex_areas = _np.zeros(grid.number_of_vertices, dtype="float64")

        vertex_used = _np.full(grid.number_of_vertices, False)

        for element_index in self.space.support_elements:
            local_values = self.evaluate(element_index, local_coordinates)
            for i in range(3):
                index = grid.elements[i, element_index]
                vertex_used[index] = True
                element_area = grid.volumes[element_index]
                vertex_areas[index] += element_area
                values[:, index] += local_values[:, i] * element_area

        values[:, vertex_used] /= vertex_areas[vertex_used]
        return values

    def integrate(self):
        """Integrate grid function over a grid."""
        from bempp_cl.api.integration.triangle_gauss import rule

        points, weights = rule(self._parameters.quadrature.regular)

        return _integrate(
            self.grid_coefficients,
            self.space.grid.data("double"),
            self.space.support_elements,
            self.space.local2global,
            self.space.local_multipliers,
            self.space.normal_multipliers,
            self.space.numba_evaluate,
            self.space.shapeset.evaluate,
            points,
            weights,
            self.component_count,
            self.space.number_of_shape_functions,
        )

    def l2_norm(self):
        """Calculate the L^2 norm of the function."""
        import numpy as np

        # L2-Norm on the whole space
        mass = self.space.mass_matrix()
        vec = self.coefficients
        return np.sqrt(np.abs(vec.conjugate().T.dot(mass.dot(vec))))

    def __add__(self, other):
        """Add two grid functions."""
        if self.space != other.space:
            raise ValueError("Spaces are not identical.")

        if self.representation == "dual" and other.representation == "dual":
            if self.dual_space == other.dual_space:
                return GridFunction(
                    self.space,
                    projections=self.projections() + other.projections(),
                    dual_space=self.dual_space,
                )

        return GridFunction(self.space, coefficients=self.coefficients + other.coefficients)

    def __mul__(self, alpha):
        """Multiply a grid function."""
        import numpy as np

        if np.isscalar(alpha):
            if self.representation == "dual":
                return GridFunction(
                    self.space,
                    projections=alpha * self._projections,
                    dual_space=self.dual_space,
                    parameters=self.parameters,
                )
            else:
                return GridFunction(
                    self.space,
                    coefficients=alpha * self.coefficients,
                    parameters=self.parameters,
                )
        else:
            return NotImplemented

    def __rmul__(self, alpha):
        """Multiply a grid function."""
        import numpy as np

        if np.isscalar(alpha):
            return self * alpha
        else:
            return NotImplemented

    def __div__(self, alpha):
        """Divide a grid function."""
        return self * (1.0 / alpha)

    def __truediv__(self, alpha):
        """Divide a grid function."""
        return self.__div__(alpha)

    def __neg__(self):
        """Negate a grid function."""
        return self.__mul__(-1.0)

    def __sub__(self, other):
        """Subtract grid functions."""
        if self.space != other.space:
            raise ValueError("Spaces are not identical.")

        return self + (-other)

    @classmethod
    def from_random(cls, space):
        """Create a random grid function normalized to unit norm."""
        from numpy.random import randn

        ndofs = space.global_dof_count
        fun = cls(space, coefficients=randn(ndofs))
        return fun / fun.l2_norm()

    @classmethod
    def from_ones(cls, space):
        """Create a grid function with all coefficients set to one."""
        from numpy import ones

        ndofs = space.global_dof_count

        return cls(space, coefficients=ones(ndofs))

    @classmethod
    def from_zeros(cls, space):
        """Create a grid function with all coefficients set to one."""
        from numpy import zeros

        ndofs = space.global_dof_count

        return cls(space, coefficients=zeros(ndofs))

    # @classmethod
    # def from_grid_interpolation(cls, grid, mode, callable):
    # """
    # Obtain a grid function from interpolation on the grid.

    # Parameters
    # ----------
    # grid : Bempp Grid object
    # The grid to be used
    # mode : string
    # Either "elements" or "vertices". If mode is
    # "elements" the callable is interpolated on
    # element centers. If mode is "vertices" the
    # callable is interpolated on vertices.
    # callable : callable object
    # A Python callable of the form:
    # values = callable(points). 'points' is
    # a float64 Numpy array of shape (N, 3),
    # where N is the number of interpolation points.
    # values is a scalar Numpy array of interpolation
    # values.

    # Returns a grid function of space type ("DP", 0) for elementwise
    # interpolation and ("P", 1) for vertex interpolation.

    # """
    # import bempp_cl.api

    # if mode == 'elements':
    # space = bempp_cl.api.function_space(grid, "DP", 0)
    # points = grid.centroids
    # elif mode == 'vertices':
    # space = bempp_cl.api.function_space(grid, "P", 1)
    # points = grid.vertices.T
    # else:
    # raise ValueError("'mode' must be one of 'elements' or 'vertices'.")

    # values = callable(points)
    # return bempp_cl.api.GridFunction(space, coefficients=values)


@_numba.njit
def _integrate(
    coefficients,
    grid_data,
    support_elements,
    local2global,
    local_multipliers,
    normal_multipliers,
    evaluate_on_element,
    shapeset_evaluate,
    points,
    weights,
    codomain_dimension,
    number_of_shape_functions,
):
    """Integrate a grid function over a grid."""
    result = _np.zeros(codomain_dimension, dtype=coefficients.dtype)

    for index in support_elements:
        element_vals = evaluate_on_element(
            index,
            shapeset_evaluate,
            points,
            grid_data,
            local_multipliers,
            normal_multipliers,
        )

        result += (
            _np.sum(
                _np.sum(
                    (element_vals * weights)
                    * (coefficients[local2global[index]] * local_multipliers[index]).reshape(
                        number_of_shape_functions, 1
                    ),
                    axis=-1,
                ),
                axis=-1,
            )
            * grid_data.integration_elements[index]
        )

    return result


@_numba.njit
def get_function_quadrature_information(grid_data, support_elements, normal_multipliers, quad_points):
    """Return vectorized version of quad_points, normals and domain_indices."""
    nelements = len(support_elements)
    nlocal = quad_points.shape[1]
    npoints = nlocal * nelements

    global_quad_points = _np.empty((3, npoints), dtype=_np.float64)
    global_normals = _np.empty((3, npoints), dtype=_np.float64)
    global_domain_indices = _np.empty(npoints, dtype=_np.float64)

    for index, element in enumerate(support_elements):
        global_quad_points[:, nlocal * index : nlocal * (1 + index)] = grid_data.local2global(element, quad_points)
        for local_index in range(nlocal):
            global_normals[:, nlocal * index + local_index] = grid_data.normals[element] * normal_multipliers[element]
            global_domain_indices[nlocal * index + local_index] = grid_data.domain_indices[element]

    return (global_quad_points, global_normals, global_domain_indices)


# Must be used in jit mode as fun might just be a Python callable and not numba compiled.
@_numba.njit
def _project_function(
    fun,
    grid_data,
    support_elements,
    local2global,
    local_multipliers,
    normal_multipliers,
    evaluate_on_element,
    shapeset_evaluate,
    points,
    weights,
    codomain_dimension,
    projections,
    function_parameters,
):
    """Project a Numba callable onto a grid."""
    npoints = points.shape[1]
    global_points = _np.empty((3, npoints), dtype=_np.float64)
    fvalues = _np.empty((codomain_dimension, npoints), dtype=projections.dtype)
    fun_result = _np.empty(codomain_dimension, dtype=projections.dtype)
    point = _np.empty(3, dtype=_np.float64)

    for index in support_elements:
        element_vals = evaluate_on_element(
            index,
            shapeset_evaluate,
            points,
            grid_data,
            local_multipliers,
            normal_multipliers,
        )

        for j in range(3):
            global_points[j] = (
                (1.0 - points[0] - points[1]) * grid_data.vertices[j, grid_data.elements[0, index]]
                + points[0] * grid_data.vertices[j, grid_data.elements[1, index]]
                + points[1] * grid_data.vertices[j, grid_data.elements[2, index]]
            )

        for j in range(npoints):
            point = global_points[:, j]

            fun(
                point,
                grid_data.normals[index] * normal_multipliers[index],
                grid_data.domain_indices[index],
                fun_result,
                function_parameters,
            )
            fvalues[:, j] = fun_result

        for local_fun_index in range(element_vals.shape[1]):
            projections[local2global[index, local_fun_index]] += (
                _np.sum(_np.sum(element_vals[:, local_fun_index, :] * fvalues * weights, axis=0))
                * grid_data.integration_elements[index]
            )


@_numba.njit
def _project_function_vectorized(
    function_data,
    grid_data,
    support_elements,
    local2global,
    local_multipliers,
    normal_multipliers,
    evaluate_on_element,
    shapeset_evaluate,
    points,
    weights,
    codomain_dimension,
    projections,
    function_parameters,
):
    """Project a Numba callable onto a grid."""
    npoints = points.shape[1]

    for index, element in enumerate(support_elements):
        element_vals = evaluate_on_element(
            element,
            shapeset_evaluate,
            points,
            grid_data,
            local_multipliers,
            normal_multipliers,
        )

        for local_fun_index in range(element_vals.shape[1]):
            projections[local2global[element, local_fun_index]] += (
                _np.sum(
                    _np.sum(
                        element_vals[:, local_fun_index, :]
                        * function_data[:, index * npoints : (1 + index) * npoints]
                        * weights,
                        axis=0,
                    )
                )
                * grid_data.integration_elements[element]
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 17 08:29:39 2024

@author: tom
"""

class postProcess:
    def __init__(self):
        """
        Store transfer function of corresponding radiating elements.
        Should be set in a evaluation.plot() method call.
        
        ex:
        EV = evaluationObject
        EV.plot(evaluations=[], radiatingElement=[1, 2, 3], 
                processing=postProcessObj)
        
        -> will plot all evaluations with summed elements' TF for 1, 2 and 3.
        """
        self.TF = {}        # transfer-functions

    def addTransferFunction(self,  name, H, radiatingElement):
        """
        Add a transfer function to the TF dictionnary

        Parameters
        ----------
        name: str
            reference name of the transfer function
        H : ndarray
            transfer function (complex). Must have the same dimension as the 
            BEM/evaluation frequency axis (i.e. same frequency bins)
            
        radiatingElement: int or list of int
            radiating element of the evaluation on which to add the 
            corresponding transfer function.
        Returns
        -------
        None.

        """
        self.TF[name] = {}
        self.TF[name]["H"] = H
        self.TF[name]["radiatingElement"] = radiatingElement
        
        
        
        
        
        
        
        


"""Timing functions."""

import time as _time


def timeit(message):
    """Time a method in bempp_cl."""

    def timeit_impl(fun):
        """Return an implementation of timeit."""

        def timed_fun(*args, **kwargs):
            """Time something."""
            from bempp_cl.api import GLOBAL_PARAMETERS
            from bempp_cl.api import log

            if not GLOBAL_PARAMETERS.verbosity.extended_verbosity:
                return fun(*args, **kwargs)

            start_time = _time.time()
            res = fun(*args, **kwargs)
            end_time = _time.time()
            log(message + " : {0:.3e}s".format(end_time - start_time))
            return res

        return timed_fun

    return timeit_impl


# pylint: disable=too-few-public-methods
class Timer:
    """Context manager to measure time in bempp_cl."""

    def __init__(self):
        """Construct."""
        self.start = 0
        self.end = 0
        self.interval = 0

    def __enter__(self):
        """Enter."""
        self.start = _time.time()
        return self

    def __exit__(self, *args):
        """Exit."""
        self.end = _time.time()
        self.interval = self.end - self.start



import gmsh
import numpy as np
from copy import copy

class CAD:
    def __init__(self, file, minSize=0.0057, maxSize=0.057, scaling=0.001, meshAlgo=6):
        """
        Import a geometry to be meshed

        Parameters
        ----------
        file : str
            path to .step file.
        minSize : float, optional
            Minimum mesh size (in m). The default is 0.0057.
        maxSize : float, optional
            Maximum mesh size (in m). The default is 0.057.
        scaling : float, optional
            Scaling of geometry (Gmsh tends to get rid of units, hence mm end up in m). The default is 0.001.
        meshAlgo : int, optional
            Type of mesh. The default is 6 ().

        Returns
        -------
        None.

        """
        
        gmsh.initialize()
        gmsh.option.setNumber("General.Terminal", 1)
        gmsh.option.setNumber("Geometry.OCCScaling", scaling)
        gmsh.option.setNumber("Mesh.Algorithm", meshAlgo)
        gmsh.option.setNumber("Mesh.MeshSizeMin", minSize)
        gmsh.option.setNumber("Mesh.MeshSizeMax", maxSize)
        gmsh.option.setNumber("Mesh.Optimize", 1)
        gmsh.option.setNumber("Mesh.QualityType", 2)
        gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)

        gmsh.merge(file)
        n = gmsh.model.getDimension()
        s = gmsh.model.getEntities(n)
        l = gmsh.model.geo.addSurfaceLoop([s[i][1] for i in range(len(s))])

        self.file = file
        self.minSize  = minSize
        self.maxSize  = maxSize
        self.scaling  = scaling
        self.meshAlgo = 6
        self.entities = gmsh.model.get_entities()
        self.surface_list = get_tags_by_dimension(self.entities, 2)
        # self.ungrouped_surface = copy(self.surface_list)
        self.physical_groups = []



    def addSurfaceGroup(self, name, surface, groupNumber, meshSize=None):
        if meshSize == None:
            meshSize = self.maxSize
        else:
            meshSize = meshSize
        SURF_GROUP = {"name": name, "surface": surface, "groupNumber": groupNumber, 
                      "meshSize": meshSize, "id": "surface"}
        self.physical_groups.append(SURF_GROUP)
        return None
    
    def addVolumeGroup(self, name, volume, groupNumber, meshSize=None):
        if meshSize == None:
            meshSize = self.maxSize
        else:
            meshSize = meshSize
        VOL_GROUP = {"name": name, "volume": volume, "groupNumber": groupNumber, 
                      "meshSize": meshSize, "id":"volume"}
        self.physical_groups.append(VOL_GROUP)
        return None

    def mesh(self, filename, order=2, excludeRemaining=False, reverseNormals=False):
        # initialization of GMSH

        surface = []
        evalField = np.arange(1, len(self.physical_groups)*2, 2)
        sizeField = evalField + 1
        for i in range(len(self.physical_groups)):
            # group surfaces
            if  self.physical_groups[i]["id"] == "surface":
                surface_tmp = self.physical_groups[i]["surface"]
                groupNumber_tmp = self.physical_groups[i]["groupNumber"]
                name_tmp =  self.physical_groups[i]["name"]
                gmsh.model.geo.add_physical_group(2, surface_tmp, groupNumber_tmp, name_tmp)
                surface.append(surface_tmp)
            elif self.physical_groups[i]["id"] == "volume":
                volume_tmp = self.physical_groups[i]["volume"]
                groupNumber_tmp = self.physical_groups[i]["groupNumber"]
                name_tmp =  self.physical_groups[i]["name"]
                gmsh.model.geo.add_physical_group(3, volume_tmp,
                                                  groupNumber_tmp, name_tmp)
            
        self.ungrouped_surface = filter_surface_list(self.surface_list, surface)
        if excludeRemaining is False:
            gmsh.model.geo.add_physical_group(2, self.ungrouped_surface, -1, "enclosure")
        else:
            pass

        for i in range(len(self.physical_groups)):
            # define mesh field
            # TODO find a way to fix the size modif for volumes
            if self.physical_groups[i]["id"] == "surface":
                gmsh.model.mesh.field.add("MathEval", evalField[i])
                gmsh.model.mesh.field.set_string(evalField[i], "F", 
                                                 str(self.physical_groups[i]["meshSize"]))
                gmsh.model.mesh.field.add("Restrict", sizeField[i])
                gmsh.model.mesh.field.setNumbers(sizeField[i], "SurfacesList", 
                                                 self.physical_groups[i]["surface"])
                gmsh.model.mesh.field.setNumber(sizeField[i], "InField",evalField[i])
            else:
                pass

        gmsh.model.mesh.field.add("Min", evalField[-1]+2)
        gmsh.model.mesh.field.setNumbers(evalField[-1]+2, "FieldsList", list(sizeField))
        gmsh.model.mesh.field.setAsBackgroundMesh(evalField[-1]+2)
        
        
        # ====================================================================
        # extrude PML layer of thickness d_PML and Num_layers layers
        # gmsh.option.setNumber('Geometry.ExtrudeReturnLateralEntities', 0)
        # e = gmsh.model.geo.extrudeBoundaryLayer([(2, 11)], 
        #                                         [5], [0.025],
        #                                         False)

        # # retrieve tags from PML interface (bottom surf), PML subvolumes
        # bottom_surf  = [s[1] for s in [(2, 11)]]
        # pml_entities = [s for s in e if s[0] == 3]
        # pml_volumes  = [s[1] for s in pml_entities]
        
        
        # # set PML volume and PML interface physical groups, generate the entire 3D mesh
        # # and export it in .msh file
        # gmsh.model.addPhysicalGroup(3, pml_volumes, 20, "PML")
        # print("bottom_surf: ", bottom_surf)
        # gmsh.model.addPhysicalGroup(2, bottom_surf, 21, "pml_int")
        # ====================================================================
        
        gmsh.model.geo.synchronize()
        gmsh.model.mesh.generate(order)

        if reverseNormals is True:
            # Get elements and nodes of the mesh
            gmsh.model.mesh.reverse()


        gmsh.write(filename + ".msh")
        gmsh.finalize()
        return None

    def export2stl(self, fileName):
        gmsh.write(fileName + ".stl")
        return None

    def mesh2tmp(self, fileName):
        for i in range(len(self.surface_list)):
            self.addSurfaceGroup(self.surface_list[i], [self.surface_list[i]], self.surface_list[i])
        gmsh.model.mesh.generate(2)
        gmsh.write(fileName + ".msh")
        return None

    def finalize(self):
        gmsh.finalize()


# Useful function
def get_tags_by_dimension(tag_list, target_dimension):
    tags = [tag for dim, tag in tag_list if dim == target_dimension]
    return tags

def filter_surface_list(surface_list, list_of_small_lists):
    exclusion_set = set().union(*list_of_small_lists)
    filtered_list = [item for item in surface_list if item not in exclusion_set]
    return filtered_list


"""The basic grid class."""

from bempp_cl.helpers import timeit as _timeit
import collections as _collections

import numba as _numba
import numpy as _np

EDGES_ID = 2
VERTICES_ID = 1

_EDGE_LOCAL = _np.array([[0, 1], [2, 0], [1, 2]])


class Grid(object):
    """The Grid class."""

    @_timeit
    def __init__(self, vertices, elements, domain_indices=None, grid_id=None, scatter=True):
        """Create a grid from a vertices and an elements array."""
        from bempp_cl.api import log
        from bempp_cl.api.utils import pool
        from bempp_cl.api.utils.helpers import create_unique_id

        self._vertices = None
        self._elements = None
        self._domain_indices = None
        self._edges = None
        self._element_edges = None
        self._edge_adjacency = None
        self._vertex_adjacency = None
        self._element_neighbors = None
        self._vertex_on_boundary = None
        self._edge_on_boundary = None
        self._edge_neighbors = None
        self._vertex_neighbors = None
        self._barycentric_grid = None
        if grid_id:
            self._id = grid_id
        else:
            self._id = create_unique_id()

        self._volumes = None
        self._normals = None
        self._jacobians = None
        self._jacobian_inverse_transposed = None
        self._diameters = None
        self._integration_elements = None
        self._centroids = None

        self._device_interfaces = {}

        self._element_to_vertex_matrix = None
        self._element_to_element_matrix = None

        self._normalize_and_assign_input(vertices, elements, domain_indices)
        self._enumerate_edges()

        self._get_element_adjacency_for_edges_and_vertices()
        self._compute_geometric_quantities()

        self._compute_boundary_information()

        self._compute_edge_neighbors()
        self._compute_vertex_neighbors()

        self._grid_data_double = GridDataDouble(
            self._vertices,
            self._elements,
            self._edges,
            self._element_edges,
            self._volumes,
            self._normals,
            self._jacobians,
            self._jacobian_inverse_transposed,
            self._diameters,
            self._integration_elements,
            self._centroids,
            self._domain_indices,
            self._vertex_on_boundary,
            self._element_neighbors.indices,
            self._element_neighbors.indexptr,
        )

        self._grid_data_single = GridDataFloat(
            self._vertices.astype("float32"),
            self._elements,
            self._edges,
            self._element_edges,
            self._volumes.astype("float32"),
            self._normals.astype("float32"),
            self._jacobians.astype("float32"),
            self._jacobian_inverse_transposed.astype("float32"),
            self._diameters.astype("float32"),
            self._integration_elements.astype("float32"),
            self._centroids.astype("float32"),
            self._domain_indices,
            self._vertex_on_boundary,
            self._element_neighbors.indices,
            self._element_neighbors.indexptr,
        )

        self._is_scattered = False

        if scatter and pool.is_initialised() and not pool.is_worker():
            self._scatter()
        if not pool.is_worker():
            log(
                (
                    f"Created grid with id {self.id}. Elements: {self.number_of_elements}. "
                    + f"Edges: {self.number_of_edges}. Vertices: {self.number_of_vertices}"
                )
            )

    @property
    def vertex_adjacency(self):
        """
        Vertex adjacency information.

        Returns a matrix with 4 rows. Each column has the entries e0,
        e1, ind0, ind1, which means that element e0 is connected to
        element e1 via local vertex index ind0 in e0 and ind1 in e1.
        Only returnes connectivity via a single vertex. For
        connectivity via edges see edge_adjacency.

        """
        return self._vertex_adjacency

    @property
    def edge_adjacency(self):
        """
        Edge adjacency information.

        Returns a matrix with 6 rows. Each column has the entries e0,
        e1, v00, v01, v11, v12, which means that element e0 is
        connected to element e1. Vertex v00 in element e0 is
        identical to vertex v11 in element e1, and vertex v01 in
        element 0 is identical to vertex v12 in element e1.
        """
        return self._edge_adjacency

    @property
    def element_to_vertex_matrix(self):
        """Return the matrix mapping vertices to elements."""
        return self._element_to_vertex_matrix

    @property
    def element_to_element_matrix(self):
        """
        Return element to element matrix.

        If entry (i,j) has the value n > 0, element i
        and element j are connected via n vertices.
        """
        return self._element_to_element_matrix

    @property
    def element_neighbors(self):
        """
        Return named tuple (indices, indexptr).

        The neighbors of element i are given as
        element_neighbors.indices[
            element_neighbors.indptr[i] : element_neighbors.indptr[i +1]].
        Note that the element i is contained in the list of neighbors.

        """
        return self._element_neighbors

    @property
    def number_of_vertices(self):
        """Return number of vertices."""
        return self._vertices.shape[1]

    @property
    def number_of_edges(self):
        """Return number of edges."""
        return self._edges.shape[1]

    @property
    def number_of_elements(self):
        """Return number of elements."""
        return self._elements.shape[1]

    @property
    def vertices(self):
        """Return vertices."""
        return self._vertices

    @property
    def elements(self):
        """Return elements."""
        return self._elements

    @property
    def edges(self):
        """Return edges."""
        return self._edges

    @property
    def centroids(self):
        """Return the centroids of the elements."""
        return self._centroids

    @property
    def domain_indices(self):
        """Return domain indices."""
        return self._domain_indices

    @property
    def element_edges(self):
        """
        Return an array of edge indices for each element.

        element_edges[i, j] is the index of the ith edge
        in the jth element.

        """
        return self._element_edges

    @property
    def device_interfaces(self):
        """Return the dictionary of device interfaces for the grid."""
        return self._device_interfaces

    @property
    def as_array(self):
        """
        Convert the grid to an array.

        For a grid with N elements returns a 1d array with
        9 * N entries. The three nodes for element with index e
        can be found in [9 * e, 9 * (e + 1)].

        """
        return self.vertices.T[self.elements.flatten(order="F"), :].flatten(order="C")

    @property
    def bounding_box(self):
        """
        Return the bounding box for the grid.

        The bounding box is a 3x2 array box such that
        box[:, 0] contains (xmin, ymin, zmin) and box[:, 1]
        contains (xmax, ymax, zmax).

        """
        box = _np.empty((3, 2), dtype="float64")
        box[:, 0] = _np.min(self.vertices, axis=1)
        box[:, 1] = _np.max(self.vertices, axis=1)

        return box

    @property
    def volumes(self):
        """Return element volumes."""
        return self._volumes

    @property
    def diameters(self):
        """Return element diameters."""
        return self._diameters

    @property
    def maximum_element_diameter(self):
        """Return the maximum element diameter."""
        return _np.max(self.diameters)

    @property
    def minimum_element_diameter(self):
        """Return the maximum element diameter."""
        return _np.min(self.diameters)

    @property
    def normals(self):
        """Return normals."""
        return self._normals

    @property
    def jacobians(self):
        """Return Jacobians."""
        return self._jacobians

    @property
    def integration_elements(self):
        """Return integration elements."""
        return self._integration_elements

    @property
    def jacobian_inverse_transposed(self):
        """Return the jacobian inverse transposed."""
        return self._jacobian_inverse_transposed

    @property
    def vertex_on_boundary(self):
        """Return vertex boundary information."""
        return self._vertex_on_boundary

    @property
    def edge_on_boundary(self):
        """Return edge boundary information."""
        return self._edge_on_boundary

    @property
    def edge_neighbors(self):
        """Return for each edge the list of neighboring elements.."""
        return self._edge_neighbors

    def data(self, precision="double"):
        """Return Numba container with all relevant grid data."""
        if precision == "double":
            return self._grid_data_double
        elif precision == "single":
            return self._grid_data_single
        else:
            raise ValueError("precision must be one of 'single', 'double'")

    @property
    def vertex_neighbors(self):
        """Return for each vertex the list of neighboring elements."""
        return self._vertex_neighbors

    @property
    def barycentric_refinement(self):
        """Return the barycentric refinement of this grid."""
        if self._barycentric_grid is None:
            self._barycentric_grid = barycentric_refinement(self)
        return self._barycentric_grid

    @property
    def id(self):
        """Return a unique id for the grid."""
        return self._id

    def _scatter(self):
        """Initialise the grid on all workers."""
        from bempp_cl.api.utils import pool

        array_proxies = pool.to_buffer(self.vertices, self.elements, self.domain_indices)

        pool.execute(_grid_scatter_worker, self.id, array_proxies)
        self._is_scattered = True

    def entity_count(self, codim):
        """Return the number of entities of given codimension."""
        if codim == 0:
            return self.number_of_elements
        if codim == 1:
            return self.number_of_edges
        if codim == 2:
            return self.number_of_vertices

        raise ValueError("codim must be one of 0, 1, or 2.")

    def plot(self):
        """Plot the grid."""
        from bempp_cl.api.external.viewers import visualize

        visualize(self)

    def get_element(self, index):
        """Return element with a given index."""
        return Element(self, index)

    def entity_iterator(self, codim):
        """Return an iterator for a given codim."""

        def element_iterator():
            """Iterate over elements."""
            for index in range(self.number_of_elements):
                yield Element(self, index)

        def vertex_iterator():
            """Iterate over vertices."""
            for index in range(self.number_of_vertices):
                yield Vertex(self, index)

        def edge_iterator():
            """Iterate over edges."""
            for index in range(self.number_of_edges):
                yield Edge(self, index)

        if codim not in [0, 1, 2]:
            raise ValueError("codim must be one of 0, 1, or 2.")

        if codim == 0:
            iterator = element_iterator()
        elif codim == 1:
            iterator = edge_iterator()
        elif codim == 2:
            iterator = vertex_iterator()

        return iterator

    def map_to_point_cloud(self, order=None, local_points=None, precision="double"):
        """
        Return a point cloud representation of the grid on quadratur points.

        Return a representation of the grid as a point cloud using points on
        each element either defined through a triangle Gauss qudrature order
        or by directly specifying an array of local points.

        Parameters
        ----------
        order : Integer
            Optional parameter. Specify a quadrature order for the point
            cloud generation.
        local_points: Numpy array
            A 2 x N array of N points in local reference coordinates that specify
            the points to use for each triangle.
        precision: String
            Either 'single' or 'double'.

        If neither order nor local_points is specified the quadrature order is
        obtained from the global parameters.

        Returns a M x 3 array of M points that represent the grid on the specified
        points.

        """
        import bempp_cl.api
        from bempp_cl.api.integration.triangle_gauss import rule

        if local_points is None:
            if order is None:
                order = bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular
            local_points, _ = rule(order)

        return grid_to_points(self.data("double"), local_points)

    def refine(self):
        """Return a new grid with all elements refined."""
        new_number_of_vertices = self.number_of_edges + self.number_of_vertices

        new_vertices = _np.empty((3, new_number_of_vertices), dtype="float64", order="F")

        new_vertices[:, : self.number_of_vertices] = self.vertices

        # Each edge midpoint forms a new vertex.
        new_vertices[:, self.number_of_vertices :] = 0.5 * (
            self.vertices[:, self.edges[0, :]] + self.vertices[:, self.edges[1, :]]
        )

        new_elements = _np.empty((3, 4 * self.number_of_elements), order="F", dtype="uint32")

        new_domain_indices = _np.repeat(self.domain_indices, 4)

        for index, elem in enumerate(self.elements.T):
            vertex0 = elem[0]
            vertex1 = elem[1]
            vertex2 = elem[2]
            vertex01 = self.element_edges[0, index] + self.number_of_vertices
            vertex20 = self.element_edges[1, index] + self.number_of_vertices
            vertex12 = self.element_edges[2, index] + self.number_of_vertices

            new_elements[:, 4 * index] = [vertex0, vertex01, vertex20]

            new_elements[:, 4 * index + 1] = [vertex01, vertex1, vertex12]

            new_elements[:, 4 * index + 2] = [vertex12, vertex2, vertex20]

            new_elements[:, 4 * index + 3] = [vertex01, vertex12, vertex20]

        return Grid(new_vertices, new_elements, new_domain_indices)

    def _compute_vertex_neighbors(self):
        """Return all elements adjacent to a given vertex."""
        from bempp_cl.helpers import IndexList

        # self._vertex_neighbors = [None for _ in range(self.number_of_vertices)]

        indptr = self.element_to_vertex_matrix.indptr
        indices = self.element_to_vertex_matrix.indices
        self._vertex_neighbors = IndexList(indices, indptr)
        # for index in range(self.number_of_vertices):
        #    self._vertex_neighbors[index] = indices[indptr[index] : indptr[index + 1]]

    def _normalize_and_assign_input(self, vertices, elements, domain_indices):
        """Convert input into the right form."""
        from bempp_cl.api.utils.helpers import align_array

        if domain_indices is None:
            domain_indices = _np.zeros(elements.shape[1], dtype="uint32")

        self._vertices = align_array(vertices, "float64", "F")
        self._elements = align_array(elements, "uint32", "F")
        self._domain_indices = align_array(domain_indices, "uint32", "F")

    def _enumerate_edges(self):
        """
        Enumerate all edges in a given grid.

        Assigns a tuple (edges, element_edges) to
        self._edges and self._element_edges.
        element_edges is an array a such that a[i, j] is the
        index of the ith edge in the jth elements, and edges
        is a 2 x nedges array such that the jth column stores the
        two nodes associated with the jth edge.

        """
        # The following would be better defined inside the njitted routiine.
        # But Numba then throws an error that it cannot find the UniTuple type.
        edge_tuple_to_index = _numba.typed.Dict.empty(
            key_type=_numba.types.containers.UniTuple(_numba.types.int64, 2),
            value_type=_numba.types.int64,
        )

        self._edges, self._element_edges = _numba_enumerate_edges(self._elements, edge_tuple_to_index)

    def _get_element_adjacency_for_edges_and_vertices(self):
        """Get element adjacency.

        The array edge_adjacency has 6 rows, such that for index j the
        element edge_adjacency[0, j] is connected with element
        edge_adjacency[1, j] via the vertices edge_adjacency[2:4, j]
        in the first element and the vertices edge_adjacency[4:6, j]
        in the second element. The vertex numbers here are local
        numbers (0, 1 or 2).

        The array vertex_adjacency has 4 rows, such that for index j the
        element vertex_adjacency[0, j] is connected with
        vertex_adjacency[1, j] via the vertex vertex_adjacency[2, j]
        in the first element and the vertex vertex_adjacency[3, j]
        in the second element. The vertex numbers here are local numbers
        (0, 1 or 2).

        """
        from bempp_cl.helpers import IndexList

        self._element_to_vertex_matrix = get_element_to_vertex_matrix(self._vertices, self._elements)

        elem_to_elem_matrix = get_element_to_element_matrix(self._vertices, self._elements)

        self._element_to_element_matrix = elem_to_elem_matrix

        elements1, elements2, nvertices = _get_element_to_element_vertex_count(elem_to_elem_matrix)

        vertex_connected_elements1, vertex_connected_elements2 = _element_filter(
            elements1, elements2, nvertices, VERTICES_ID
        )

        edge_connected_elements1, edge_connected_elements2 = _element_filter(elements1, elements2, nvertices, EDGES_ID)

        self._vertex_adjacency = _find_vertex_adjacency(
            self._elements, vertex_connected_elements1, vertex_connected_elements2
        )

        self._edge_adjacency = _find_edge_adjacency(self._elements, edge_connected_elements1, edge_connected_elements2)

        self._element_neighbors = IndexList(elem_to_elem_matrix.indices, elem_to_elem_matrix.indptr)

    def _compute_geometric_quantities(self):
        """Compute geometric quantities for the grid."""
        element_vertices = self.vertices.T[self.elements.flatten(order="F")]
        indexptr = 3 * _np.arange(self.number_of_elements)
        indices = _np.repeat(indexptr, 2) + _np.tile([1, 2], self.number_of_elements)

        centroids = 1.0 / 3 * _np.sum(_np.reshape(element_vertices, (self.number_of_elements, 3, 3)), axis=1)

        jacobians = (element_vertices - _np.repeat(element_vertices[::3], 3, axis=0))[indices]

        normal_directions = _np.cross(jacobians[::2], jacobians[1::2], axis=1)
        normal_direction_norms = _np.linalg.norm(normal_directions, axis=1)
        normals = normal_directions / _np.expand_dims(normal_direction_norms, 1)

        volumes = 0.5 * normal_direction_norms

        jacobian_diff = jacobians[::2] - jacobians[1::2]
        diff_norms = _np.linalg.norm(jacobian_diff, axis=1)
        jac_vector_norms = _np.linalg.norm(jacobians, axis=1)

        diameters = jac_vector_norms[::2] * jac_vector_norms[1::2] * diff_norms / normal_direction_norms

        self._volumes = volumes
        self._normals = normals
        self._jacobians = _np.swapaxes(_np.reshape(jacobians, (self.number_of_elements, 2, 3)), 1, 2)
        self._diameters = diameters
        self._centroids = centroids

        jac_transpose_jac = _np.empty((self.number_of_elements, 2, 2), dtype="float64")
        for index in range(self.number_of_elements):
            jac_transpose_jac[index] = self.jacobians[index].T.dot(self.jacobians[index])
        self._integration_elements = _np.sqrt(_np.linalg.det(jac_transpose_jac))

        jac_transpose_jac_inv = _np.linalg.inv(jac_transpose_jac)

        self._jacobian_inverse_transposed = _np.empty((self.number_of_elements, 3, 2), dtype="float64")

        for index in range(self.number_of_elements):
            self._jacobian_inverse_transposed[index] = self.jacobians[index].dot(jac_transpose_jac_inv[index])

    def _compute_boundary_information(self):
        """
        Return a boolean array with boundary information.

        Computes arr0, arr1 such that arr0[j] is True if
        vertex j lies on the boundary and arr1[i] is True if edge
        i lies on the boundary.
        """
        from scipy.sparse import csr_matrix

        element_edges = self.element_edges

        number_of_elements = self.number_of_elements
        number_of_edges = self.number_of_edges
        number_of_vertices = self.number_of_vertices
        edge_indices = _np.ravel(element_edges, order="F")
        repeated_element_indices = _np.repeat(_np.arange(number_of_elements), 3)
        data = _np.ones(3 * number_of_elements, dtype="uint32")

        element_to_edge = csr_matrix(
            (data, (repeated_element_indices, edge_indices)),
            shape=(number_of_elements, number_of_edges),
        )

        edge_to_edge = element_to_edge.T.dot(element_to_edge)
        arr1 = edge_to_edge.diagonal() == 1
        arr0 = _np.full(number_of_vertices, False)

        for boundary_edge_index in _np.flatnonzero(arr1):
            arr0[self.edges[:, boundary_edge_index]] = True

        self._vertex_on_boundary = arr0
        self._edge_on_boundary = arr1

    def _compute_edge_neighbors(self):
        """Get the neighbors of each edge."""
        edge_neighbors = [[] for _ in range(self.number_of_edges)]

        for element_index in range(self.number_of_elements):
            for local_index in range(3):
                edge_neighbors[self.element_edges[local_index, element_index]].append(element_index)
        self._edge_neighbors = [tuple(elem) for elem in edge_neighbors]


@_numba.experimental.jitclass(
    [
        ("vertices", _numba.float64[:, :]),
        ("elements", _numba.uint32[:, :]),
        ("edges", _numba.uint32[:, :]),
        ("element_edges", _numba.uint32[:, :]),
        ("volumes", _numba.float64[:]),
        ("normals", _numba.float64[:, :]),
        ("jacobians", _numba.float64[:, :, :]),
        ("jac_inv_trans", _numba.float64[:, :, :]),
        ("diameters", _numba.float64[:]),
        ("integration_elements", _numba.float64[:]),
        ("centroids", _numba.float64[:, :]),
        ("domain_indices", _numba.uint32[:]),
        ("vertex_on_boundary", _numba.boolean[:]),
        ("element_neighbor_indices", _numba.uint32[:]),
        ("element_neighbor_indexptr", _numba.uint32[:]),
    ]
)
class GridDataDouble(object):
    """A Numba container class for the grid data."""

    def __init__(
        self,
        vertices,
        elements,
        edges,
        element_edges,
        volumes,
        normals,
        jacobians,
        jac_inv_trans,
        diameters,
        integration_elements,
        centroids,
        domain_indices,
        vertex_on_boundary,
        element_neighbor_indices,
        element_neighbor_indexptr,
    ):
        """Create a GridDataDouble."""
        self.vertices = vertices
        self.elements = elements
        self.edges = edges
        self.element_edges = element_edges
        self.volumes = volumes
        self.normals = normals
        self.jacobians = jacobians
        self.jac_inv_trans = jac_inv_trans
        self.diameters = diameters
        self.integration_elements = integration_elements
        self.centroids = centroids
        self.domain_indices = domain_indices
        self.vertex_on_boundary = vertex_on_boundary
        self.element_neighbor_indices = element_neighbor_indices
        self.element_neighbor_indexptr = element_neighbor_indexptr

    def local2global(self, elem_index, local_coords):
        """Map local to global coordinates."""
        return _np.expand_dims(self.vertices[:, self.elements[0, elem_index]], 1) + self.jacobians[elem_index].dot(
            local_coords
        )


@_numba.experimental.jitclass(
    [
        ("vertices", _numba.float32[:, :]),
        ("elements", _numba.uint32[:, :]),
        ("edges", _numba.uint32[:, :]),
        ("element_edges", _numba.uint32[:, :]),
        ("volumes", _numba.float32[:]),
        ("normals", _numba.float32[:, :]),
        ("jacobians", _numba.float32[:, :, :]),
        ("jac_inv_trans", _numba.float32[:, :, :]),
        ("diameters", _numba.float32[:]),
        ("integration_elements", _numba.float32[:]),
        ("centroids", _numba.float32[:, :]),
        ("domain_indices", _numba.uint32[:]),
        ("vertex_on_boundary", _numba.boolean[:]),
        ("element_neighbor_indices", _numba.uint32[:]),
        ("element_neighbor_indexptr", _numba.uint32[:]),
    ]
)
class GridDataFloat(object):
    """A Numba container class for the grid data."""

    def __init__(
        self,
        vertices,
        elements,
        edges,
        element_edges,
        volumes,
        normals,
        jacobians,
        jac_inv_trans,
        diameters,
        integration_elements,
        centroids,
        domain_indices,
        vertex_on_boundary,
        element_neighbor_indices,
        element_neighbor_indexptr,
    ):
        """Create a GridDataFloat."""
        self.vertices = vertices
        self.elements = elements
        self.edges = edges
        self.element_edges = element_edges
        self.volumes = volumes
        self.normals = normals
        self.jacobians = jacobians
        self.jac_inv_trans = jac_inv_trans
        self.diameters = diameters
        self.integration_elements = integration_elements
        self.centroids = centroids
        self.domain_indices = domain_indices
        self.vertex_on_boundary = vertex_on_boundary
        self.element_neighbor_indices = element_neighbor_indices
        self.element_neighbor_indexptr = element_neighbor_indexptr

    def local2global(self, elem_index, local_coords):
        """Map local to global coordinates."""
        return _np.expand_dims(self.vertices[:, self.elements[0, elem_index]], 1) + self.jacobians[elem_index].dot(
            local_coords
        )


class ElementGeometry(object):
    """Provides geometry information for an element."""

    def __init__(self, grid, index):
        """Initialize geometry wth a 3x3 array of corners."""
        self._grid = grid
        self._index = index

    @property
    def corners(self):
        """Return corners."""
        return self._grid.vertices[:, self._grid.elements[:, self._index]]

    @property
    def jacobian(self):
        """Return jacobian."""
        return self._grid.jacobians[self._index]

    @property
    def integration_element(self):
        """Return integration element."""
        return self._grid.integration_elements[self._index]

    @property
    def jacobian_inverse_transposed(self):
        """Return Jacobian inverse transposed."""
        return self._grid.jacobian_inverse_transposed[self._index]

    @property
    def normal(self):
        """Return normal."""
        return self._grid.normals[self._index]

    @property
    def volume(self):
        """Return volume."""
        return self._grid.volumes[self._index]

    @property
    def diameter(self):
        """Return the diameter of the circumcircle."""
        return self._grid.diameters[self._index]

    @property
    def centroid(self):
        """Return the centroid of the element."""
        return self._grid.centroids[self._index]

    def local2global(self, points):
        """Map points in local coordinates to global."""
        return _np.expand_dims(self.corners[:, 0], 1) + self.jacobian @ points


class Element(object):
    """Provides a view onto an element of the grid."""

    def __init__(self, grid, index):
        self._grid = grid
        self._index = index

    @property
    def index(self):
        """Index of the element."""
        return self._index

    @property
    def grid(self):
        """Associated grid."""
        return self._grid

    @property
    def geometry(self):
        """Return geometry."""
        grid = self._grid
        return ElementGeometry(grid, self.index)

    @property
    def domain_index(self):
        """Return the domain index."""
        return self._grid.domain_indices[self.index]

    def sub_entity_iterator(self, codim):
        """Return iterator over subentitites."""

        def edge_iterator():
            """Iterate over edges."""
            for index in self._grid.element_edges[:, self.index]:
                yield Edge(self._grid, index)

        def vertex_iterator():
            """Iterate over vertices."""
            for index in self._grid.elements[:, self.index]:
                yield Vertex(self._grid, index)

        if codim not in [1, 2]:
            raise ValueError("codim must be 1 (for edges) or 2 (for vertices)")

        if codim == 1:
            iterator = edge_iterator()

        if codim == 2:
            iterator = vertex_iterator()

        return iterator

    def __eq__(self, other):
        """Check if elements are equal."""
        if isinstance(other, Element):
            if other.grid == self.grid and other.index == self.index:
                return True
        return False


VertexGeometry = _collections.namedtuple("VertexGeometry", "corners")


class Vertex(object):
    """Provides a view onto a vertex of the grid."""

    def __init__(self, grid, index):
        """Create a vertex."""
        self._grid = grid
        self._index = index

    @property
    def index(self):
        """Index of the vertex."""
        return self._index

    @property
    def geometry(self):
        """Return geometry."""
        return VertexGeometry(self._grid.vertices[:, self.index].reshape(3, 1))


class EdgeGeometry(object):
    """Implementation of a geometry for edges."""

    def __init__(self, corners):
        """Create edge geometry."""
        self._corners = corners
        self._volume = _np.linalg.norm(corners[:, 1] - corners[:, 0])

    @property
    def corners(self):
        """Return the corners."""
        return self._corners

    @property
    def volume(self):
        """Return length of the edge."""
        return self._volume


class Edge(object):
    """Provides a view onto an edge of the grid."""

    def __init__(self, grid, index):
        """Create an edge."""
        self._grid = grid
        self._index = index

    @property
    def index(self):
        """Return the index of the edge."""
        return self._index

    @property
    def geometry(self):
        """Return geometry."""
        grid = self._grid
        return EdgeGeometry(grid.vertices[:, grid.edges[:, self.index]])


def get_element_to_vertex_matrix(vertices, elements):
    """Return the sparse matrix mapping vertices to elements."""
    from scipy.sparse import csr_matrix

    number_of_elements = elements.shape[1]
    number_of_vertices = vertices.shape[1]
    vertex_indices = _np.ravel(elements, order="F")
    vertex_element_indices = _np.repeat(_np.arange(number_of_elements), 3)
    data = _np.ones(len(vertex_indices), dtype="uint32")

    return csr_matrix(
        (data, (vertex_indices, vertex_element_indices)),
        shape=(number_of_vertices, number_of_elements),
        dtype="uint32",
    )


def get_element_to_element_matrix(vertices, elements):
    """
    Return element to element matrix.

    If entry (i,j) has the value n > 0, element i
    and element j are connected via n vertices.

    """
    element_to_vertex = get_element_to_vertex_matrix(vertices, elements)
    return element_to_vertex.T.dot(element_to_vertex)


@_numba.njit(locals={"index": _numba.types.int32})
def _compare_array_to_value(array, val):
    """
    Return i such that array[i] == val.

    If val not found return -1
    """
    for index, elem in enumerate(array):
        if elem == val:
            return index
    return -1


@_numba.njit(
    locals={
        "index1": _numba.types.int32,
        "index2": _numba.types.int32,
        "full_index1": _numba.types.int32,
    }
)
def _find_first_common_array_index_pair_from_position(array1, array2, start=0):
    """
    Return first index pair (i, j) such that array1[i] = array2[j].

    Assumes that one index pair satisfying the equality
    always exists. Method checks in array1 from position start
    onwards.
    """
    for index1 in range(len(array1[start:])):
        full_index1 = index1 + start
        index2 = _compare_array_to_value(array2, array1[full_index1])
        if index2 != -1:
            return (full_index1, index2)
    raise ValueError("Could not find a common index pair.")


@_numba.njit(locals={"offset": _numba.types.int32})
def _find_two_common_array_index_pairs(array1, array2):
    """Return two index pairs (i, j) such that array1[i] = array2[j]."""
    offset = 0
    index_pairs = _np.empty((2, 2), dtype=_np.int32)
    index_pairs[:, 0] = _find_first_common_array_index_pair_from_position(array1, array2, offset)
    offset = index_pairs[0, 0] + 1  # Next search starts behind found pair
    index_pairs[:, 1] = _find_first_common_array_index_pair_from_position(array1, array2, offset)
    return index_pairs


@_numba.njit()
def _get_shared_vertex_information_for_two_elements(elements, elem0, elem1):
    """
    Return tuple (i, j).

    The tuple has the property elements[i, elem0] == elements[j, elem1]
    """
    i, j = _find_first_common_array_index_pair_from_position(elements[:, elem0], elements[:, elem1])
    return (i, j)


@_numba.njit()
def _get_shared_edge_information_for_two_elements(elements, elem0, elem1):
    """
    Return 2x2 array of int32 indices.

    Each column in the return indices as a pair (i, j) such that
    elements[i, elem0] = elements[j, elem1]

    """
    index_pairs = _find_two_common_array_index_pairs(elements[:, elem0], elements[:, elem1])

    # Ensure that order of indices is the same as Bempp 3

    if index_pairs[1, 1] < index_pairs[1, 0]:
        for i in range(2):
            tmp = index_pairs[i, 0]
            index_pairs[i, 0] = index_pairs[i, 1]
            index_pairs[i, 1] = tmp

    return index_pairs


@_numba.njit()
def _find_vertex_adjacency(elements, test_indices, trial_indices):
    """
    Return for element pairs the vertex adjacency.

    The return array vertex_adjacency has 4 rows, such that for index j
    the element vertex_adjacency[0, j] is connected with
    vertex_adjacency[1, j] via the vertex vertex_adjacency[2, j] in
    the first element and the vertex vertex_adjacency[3, j] in the
    second element. The vertex numbers here are local
    numbers (0, 1 or 2).

    """
    number_of_indices = len(test_indices)
    adjacency = _np.zeros((4, number_of_indices), dtype=_np.int32)

    for index in range(number_of_indices):
        # Now find the position of the shared vertex
        test_index = test_indices[index]
        trial_index = trial_indices[index]
        i, j = _get_shared_vertex_information_for_two_elements(elements, test_index, trial_index)
        adjacency[:, index] = (test_index, trial_index, i, j)

    return adjacency


@_numba.njit()
def _find_edge_adjacency(elements, elem0_indices, elem1_indices):
    """
    Return for element pairs the edge adjacency.

    The return array edge_adjacency has 6 rows, such that for index
    j the element edge_adjacency[0, j] is connected with
    edge_adjacency[1, j] via the two vertices edge_adjacency[2:4, j]
    in the first element and the vertices edge_adjacency[4:6, j]
    in the second element. The vertex numbers here are local
    numbers (0, 1 or 2).

    """
    number_of_indices = len(elem0_indices)

    adjacency = _np.zeros((6, number_of_indices), dtype=_np.int32)

    for index in range(number_of_indices):
        elem0 = elem0_indices[index]
        elem1 = elem1_indices[index]
        index_pairs = _get_shared_edge_information_for_two_elements(elements, elem0, elem1)
        adjacency[0, index] = elem0
        adjacency[1, index] = elem1
        adjacency[2:, index] = index_pairs.flatten()

    return adjacency


def _get_element_to_element_vertex_count(element_to_element_matrix):
    """
    Return a tuple of arrays (elements1, elements2, nvertices).

    The element elements1[i] is connected with elements2[i] via
    nvertices[i] vertices.

    """
    coo_matrix = element_to_element_matrix.tocoo()
    elements1 = coo_matrix.row
    elements2 = coo_matrix.col
    nvertices = coo_matrix.data

    return (elements1, elements2, nvertices)


def _element_filter(elements1, elements2, nvertices, filter_type):
    """
    Return element pairs according to a filter condition.

    Takes an array (elements1, elements2, nvertices)
    such that elements1[i] and elements2[i] are connected
    via nvertices[i] vertices and returns a tuple
    (new_elem1, new_elem2) of all element pairs connected via
    vertices (filter_type=VERTICES) or edges (filter_type=EDGES).

    """
    # Elements connected via edges share two vertices
    filtered_indices = _np.argwhere(nvertices == filter_type).flatten()
    return (elements1[filtered_indices], elements2[filtered_indices])


@_numba.njit()
def _sort_values(val1, val2):
    """Return a tuple with the input values sorted."""
    if val1 > val2:
        val1, val2 = val2, val1
    return val1, val2


@_numba.njit()
def _vertices_from_edge_index(element, local_index):
    """
    Return the vertices associated with an edge.

    Element is 3-tupel with the vertex indices.
    Sorts the returned vertices in ascending order.

    """
    vertex0, vertex1 = element[_EDGE_LOCAL[local_index]]
    return _sort_values(vertex0, vertex1)


def grid_from_segments(grid, segments):
    """Return new grid from segments of existing grid."""
    element_in_new_grid = _np.full(grid.number_of_elements, False)

    for elem in range(grid.number_of_elements):
        if grid.domain_indices[elem] in segments:
            element_in_new_grid[elem] = True
    new_elements = grid.elements[:, element_in_new_grid]
    new_domain_indices = grid.domain_indices[element_in_new_grid]
    vertex_indices = list(set(new_elements.ravel()))
    new_vertices = grid.vertices[:, vertex_indices]
    new_vertex_map = -_np.ones(grid.number_of_vertices, dtype=_np.int_)
    new_vertex_map[vertex_indices] = _np.arange(len(vertex_indices))
    new_elements = new_vertex_map[new_elements.ravel()].reshape(3, -1)

    return Grid(new_vertices, new_elements, new_domain_indices)


@_numba.njit
def _create_barycentric_connectivity_array(vertices, elements, element_edges, edges, number_of_edges):
    """Return the vertices and elements of refined barycentric grid."""
    number_of_vertices = vertices.shape[1]
    number_of_elements = elements.shape[1]
    new_number_of_vertices = number_of_vertices + number_of_elements + number_of_edges
    new_vertices = _np.empty((3, new_number_of_vertices), dtype=_np.float64)
    new_elements = _np.empty((3, 6 * number_of_elements), dtype=_np.float64)

    edge_to_vertex = -_np.ones(number_of_edges)

    new_vertices[:, :number_of_vertices] = vertices

    local_vertex_ids = _np.empty(3, dtype=_np.uint32)

    for index in range(number_of_elements):
        # Create barycentric mid-point
        new_vertices[:, number_of_vertices] = 1.0 / 3 * _np.sum(vertices[:, elements[:, index]], axis=1)
        midpoint_index = number_of_vertices
        number_of_vertices += 1
        for local_index in range(3):
            edge_index = element_edges[local_index, index]
            if edge_to_vertex[edge_index] > -1:
                # Vertex already created
                local_vertex_ids[local_index] = edge_to_vertex[edge_index]
            else:
                # Vertex needs to be created
                new_vertices[:, number_of_vertices] = 0.5 * _np.sum(vertices[:, edges[:, edge_index]], axis=1)
                local_vertex_ids[local_index] = number_of_vertices
                edge_to_vertex[edge_index] = number_of_vertices
                number_of_vertices += 1
            # Have created all necessary vertices. Now create the elements.
            # New barycentric elements are created in anti-clockwise order
            # starting with the triangle at the first vertex of the triangle
            # and sharing a segment with the edge 0. The second triangle is
            # along the same edge, but adjacent to vertex 1, and so on.

        new_elements[0, 6 * index + 0] = elements[0, index]
        new_elements[1, 6 * index + 0] = local_vertex_ids[0]
        new_elements[2, 6 * index + 0] = midpoint_index

        new_elements[0, 6 * index + 1] = elements[1, index]
        new_elements[1, 6 * index + 1] = midpoint_index
        new_elements[2, 6 * index + 1] = local_vertex_ids[0]

        new_elements[0, 6 * index + 2] = elements[1, index]
        new_elements[1, 6 * index + 2] = local_vertex_ids[2]
        new_elements[2, 6 * index + 2] = midpoint_index

        new_elements[0, 6 * index + 3] = elements[2, index]
        new_elements[1, 6 * index + 3] = midpoint_index
        new_elements[2, 6 * index + 3] = local_vertex_ids[2]

        new_elements[0, 6 * index + 4] = elements[2, index]
        new_elements[1, 6 * index + 4] = local_vertex_ids[1]
        new_elements[2, 6 * index + 4] = midpoint_index

        new_elements[0, 6 * index + 5] = elements[0, index]
        new_elements[1, 6 * index + 5] = midpoint_index
        new_elements[2, 6 * index + 5] = local_vertex_ids[1]

    return new_vertices, new_elements


def barycentric_refinement(grid):
    """Return the barycentric refinement of a given grid."""
    new_vertices, new_elements = _create_barycentric_connectivity_array(
        grid.vertices,
        grid.elements,
        grid.element_edges,
        grid.edges,
        grid.number_of_edges,
    )

    return Grid(new_vertices, new_elements, _np.repeat(grid.domain_indices, 6), scatter=False)


def union(grids, domain_indices=None, swapped_normals=None, normalize_domain_indices=True):
    """
    Return the union of a given list of grids.

    Parameters
    ----------
    grids: list
        A list of grid objects.
    domain_indices : list
        Attach a list of domain indices to the new
        grid such that grid[j] received the domain
        index domain_indices[j]
    swapped_normals : list of boolean
        A list of the form [False, True, ...],
        that specifies for each grid if the normals
        should be swapped (True) or not (False). This
        is helpful if one grid is defined to be inside
        another grid.
    normalize_domain_indices : bool
        Normalize the resulting grid with domain unique indices
        between 0 and N-1, where N is the total number
        of domains across all grids, when
        domain_indices has not been provided.

    This method returns a new grid object, which is
    the union of the input grid objects.

    """
    from bempp_cl.api.grid.grid import Grid

    vertex_offset = 0
    element_offset = 0

    vertex_count = sum([grid.number_of_vertices for grid in grids])
    element_count = sum([grid.number_of_elements for grid in grids])

    vertices = _np.empty((3, vertex_count), dtype="float64")
    elements = _np.empty((3, element_count), dtype="uint32")
    all_domain_indices = _np.empty(element_count, dtype="uint32")

    def normalize_array(arr):
        """Normalize domain indices array."""
        arr = arr - arr.min()
        for i, index in enumerate(_np.unique(arr)[1:], start=1):
            arr[arr == index] = i
        return arr

    if domain_indices is None:
        if not normalize_domain_indices:
            domain_indices = [grids[0].domain_indices]
            for grid in grids[1:]:
                domain_indices.append(domain_indices[-1].max() - grid.domain_indices.min() + 1 + grid.domain_indices)
        else:
            domain_indices = [normalize_array(grids[0].domain_indices)]
            for grid in grids[1:]:
                domain_indices.append(domain_indices[-1].max() + 1 + normalize_array(grid.domain_indices))

    if swapped_normals is None:
        swapped_normals = len(grids) * [False]

    for index, grid in enumerate(grids):
        nelements = grid.number_of_elements
        nvertices = grid.number_of_vertices
        vertices[:, vertex_offset : vertex_offset + nvertices] = grid.vertices
        if swapped_normals[index]:
            current_elements = grid.elements[[0, 2, 1], :]
        else:
            current_elements = grid.elements
        elements[:, element_offset : element_offset + nelements] = current_elements + vertex_offset
        all_domain_indices[element_offset : element_offset + nelements] = domain_indices[index]
        vertex_offset += nvertices
        element_offset += nelements
    return Grid(vertices, elements, all_domain_indices)


def enumerate_vertex_adjacent_elements(grid, support_elements, swapped_normals=None):
    """
    Enumerate in anti-clockwise order all elements adjacent to all vertices in support.

    Returns a list [neighbors_0, neighbors_1, ...], where neighbors_i is a list
    [(elem_index, local_ind1, local_ind2), ...] of tuples, where elem_index is an
    element in the support that as connected with vertex i. local_ind1 and local_ind2 are
    the local indices of the two edges that are adjacent to vertex i. They are sorted in
    anti-clockwise order with respect to the natural normal directions of the elements.
    Moreover, all tuples represent elements in anti-clockwise order.
    """
    if swapped_normals is None:
        swapped_normals = []

    vertex_edges = [[] for _ in range(grid.vertices.shape[1])]

    for element_index in support_elements:
        for local_index, edge_index in enumerate(grid.element_edges[:, element_index]):
            for ind in range(2):
                vertex_edges[grid.edges[ind, edge_index]].append((element_index, local_index))

    for vertex_index, neighbors in enumerate(vertex_edges):
        # First sort by element
        if not neighbors:
            # Continue if empty
            continue
        # Now sort each list so that edges appear in anti-clockwise order according
        # to neighboring edges.

        # Swap the edges in each element so
        # that they have edges in anti-clockwise order
        locally_sorted_neighbors = []
        while neighbors:
            # Take first element in list
            elem1 = neighbors.pop()
            for index, elem2 in enumerate(neighbors):
                # Find index of next list element associated
                # with the same grid element
                if elem2[0] == elem1[0]:
                    neighbors.pop(index)
                    break
            # Check if the two edges in the found element entries
            # are in clockwise or anti-clockwise order.
            # Resort accordingly
            if grid.domain_indices[elem1[0]] in swapped_normals:
                if elem1[1] == (1 + elem2[1]) % 3:
                    locally_sorted_neighbors.append((elem1[0], elem1[1], elem2[1]))
                else:
                    locally_sorted_neighbors.append((elem1[0], elem2[1], elem1[1]))
            else:
                if elem1[1] == (1 + elem2[1]) % 3:
                    locally_sorted_neighbors.append((elem1[0], elem2[1], elem1[1]))
                else:
                    locally_sorted_neighbors.append((elem1[0], elem1[1], elem2[1]))

        # locally sorted neighbors now has triplets (elem_index, local_ind1, local_ind2) of
        # one element index and two associated edge indices that are anti-clockwise sorted.
        sorted_neighbors = []
        sorted_neighbors.append(locally_sorted_neighbors.pop())
        while locally_sorted_neighbors:
            found = False
            for index, elem in enumerate(locally_sorted_neighbors):
                # Check if element is successor of last element in sorted list
                last = sorted_neighbors[-1]
                first = sorted_neighbors[0]
                if grid.data().element_edges[elem[1], elem[0]] == grid.data().element_edges[last[2], last[0]]:
                    locally_sorted_neighbors.pop(index)
                    found = True
                    sorted_neighbors.append(elem)
                    break
                if grid.data().element_edges[elem[2], elem[0]] == grid.data().element_edges[first[1], first[0]]:
                    locally_sorted_neighbors.pop(index)
                    found = True
                    sorted_neighbors.insert(0, elem)
                    break
            if not found:
                raise Exception("Two elements seem to be connected only by a vertex, not by an edge.")

        vertex_edges[vertex_index] = sorted_neighbors

    return vertex_edges


@_numba.njit
def _numba_enumerate_edges(elements, edge_tuple_to_index):
    """
    Enumerate all edges in a given grid.

    Assigns a tuple (edges, element_edges) to
    self._edges and self._element_edges.
    element_edges is an array a such that a[i, j] is the
    index of the ith edge in the jth elements, and edges
    is a 2 x nedges array such that the jth column stores the
    two nodes associated with the jth edge.

    """
    edges = []

    number_of_elements = elements.shape[1]
    element_edges = _np.zeros((3, number_of_elements), dtype=_np.int32)

    number_of_edges = 0

    for elem_index in range(number_of_elements):
        elem = elements[:, elem_index]
        for local_index in range(3):
            edge_tuple = _vertices_from_edge_index(elem, local_index)
            if edge_tuple not in edge_tuple_to_index:
                edge_index = number_of_edges
                edge_tuple_to_index[edge_tuple] = edge_index
                edges.append(edge_tuple)
                number_of_edges += 1
            else:
                edge_index = edge_tuple_to_index[edge_tuple]
            element_edges[local_index, elem_index] = edge_index

    return _np.array(edges, dtype=_np.int32).T, element_edges


def _grid_scatter_worker(grid_id, array_proxies):
    """Assign a new grid on the worker."""
    from bempp_cl.api.utils import pool
    from bempp_cl.api.grid.grid import Grid
    from bempp_cl.api import log

    vertices, elements, domain_indices = pool.from_buffer(array_proxies)

    # if not pool.has_key(grid_id):
    if grid_id not in pool:
        pool.insert_data(
            grid_id,
            Grid(vertices.copy(), elements.copy(), domain_indices.copy(), grid_id),
        )
        log(f"Copied grid with id {grid_id} to worker {pool.get_id()}", "debug")
    else:
        log(f"Use cached grid with id {grid_id} on worker {pool.get_id()}", "debug")


@_numba.njit
def grid_to_points(grid_data, local_points):
    """
    Map a grid to an array of points.

    Returns a (N, 3) point array that stores the global vertices
    associated with the local points in each triangle.
    Points are stored in consecutive order for each element
    in the support_elements list. Hence, the returned array is of the form
    [ v_1^1, v_2^1, ..., v_M^1, v_1^2, v_2^2, ...], where
    v_i^j is the ith point in the jth element in
    the support_elements list.

    Parameters
    ----------
    grid_data : GridData
        A Bempp GridData object.
    local_points : np.ndarray
        (2, M) array of local coordinates.
    """
    number_of_elements = grid_data.elements.shape[1]
    number_of_points = local_points.shape[1]

    points = _np.empty((number_of_points * number_of_elements, 3), dtype=_np.float64)

    for elem in range(number_of_elements):
        points[number_of_points * elem : number_of_points * (1 + elem), :] = (
            _np.expand_dims(grid_data.vertices[:, grid_data.elements[0, elem]], 1)
            + grid_data.jacobians[elem].dot(local_points)
        ).T
    return points


def _get_barycentric_support(truncate_at_segment_edge, grid, bary_grid, coarse_space):
    coarse_support = _np.zeros(grid.entity_count(0), dtype=_np.bool_)
    coarse_support[coarse_space.support_elements] = True

    if not truncate_at_segment_edge:
        for global_dof_index in range(coarse_space.global_dof_count):
            local_dofs = coarse_space.global2local[global_dof_index]
            edge_index = grid.data().element_edges[local_dofs[0][1], local_dofs[0][0]]
            for v in range(2):
                vertex = grid.data().edges[v, edge_index]
                start = grid.vertex_neighbors.indexptr[vertex]
                end = grid.vertex_neighbors.indexptr[vertex + 1]
                for cell in grid.vertex_neighbors.indices[start:end]:
                    coarse_support[cell] = True

    coarse_support_elements = _np.array([i for i, j in enumerate(coarse_support) if j])
    number_of_support_elements = len(coarse_support_elements)

    bary_support_elements = 6 * _np.repeat(coarse_support_elements, 6) + _np.tile(
        _np.arange(6), number_of_support_elements
    )

    support = _np.zeros(bary_grid.number_of_elements, dtype=_np.bool_)
    support[bary_support_elements] = True

    bary_support_size = len(bary_support_elements)

    return support, bary_support_size, bary_support_elements


def _get_data_multipliers(support, bary_grid, bary_support_elements, swapped_normals, coarse_space, bary_support_size):
    bary_vertex_to_edge = enumerate_vertex_adjacent_elements(bary_grid, bary_support_elements, swapped_normals)

    edge_vectors = bary_grid.vertices[:, bary_grid.edges[0, :]] - bary_grid.vertices[:, bary_grid.edges[1, :]]

    edge_lengths = _np.linalg.norm(edge_vectors, axis=0)

    normal_multipliers = _np.repeat(coarse_space.normal_multipliers, 6)
    local2global = _np.zeros((bary_grid.number_of_elements, 3), dtype="uint32")
    local_multipliers = _np.zeros((bary_grid.number_of_elements, 3), dtype="uint32")

    local2global[support] = _np.arange(3 * bary_support_size).reshape(bary_support_size, 3)

    local_multipliers[support] = 1

    return bary_vertex_to_edge, local2global, edge_lengths, normal_multipliers, local_multipliers


def _get_barycentric_edges_associated_to_vertex(vertex_index, bary_vertex_to_edge, bary_element, bary_grid):
    """Get all the barycentryc vertices of the elements associated to a reference edge."""
    # Get barycentric elements associated to a vertex
    for ind, elem in enumerate(bary_vertex_to_edge[vertex_index]):
        if bary_element == elem[0]:
            break

    # Get all the relevant barycentric edges starting to count above ind
    num_bary_elements = len(bary_vertex_to_edge[vertex_index])
    vertex_edges = []
    for index in range(num_bary_elements):
        elem_edge_pair = bary_vertex_to_edge[vertex_index][(index + ind) % num_bary_elements]
        for n in range(1, 3):
            vertex_edges.append((elem_edge_pair[0], elem_edge_pair[n]))
    edges = []
    for vertex in vertex_edges:
        edges.append(bary_grid.data().element_edges[vertex[1]][vertex[0]])
    set_edges = set(edges)
    sorted_edges = []
    new_vertex_edges = []
    ref_edge = 0

    for el in set_edges:
        # edges.count(el) == 1 means that the edge is associated to one element,
        # therefore, it is a barycentric element on the border of a on open geometry.
        # vertex_edges[edges.index(el)][1] == 1 indicates that the barycentric element
        # on the border is the first one anti-clock-wise.
        if edges.count(el) == 1 and vertex_edges[edges.index(el)][1] == 1:
            sorted_edges.append(el)
            new_vertex_edges.append(vertex_edges[edges.index(el)])

    # If the basis function is on the border, sort the rest of the elements accordingly.
    # The basis functions on the borders do not need to remove the reference edge.
    if len(sorted_edges) > 0:
        new_vertex_edges, sorted_edges = _sort_vertex_edges(
            vertex_edges, edges, set_edges, sorted_edges, new_vertex_edges, 1
        )
        sorted_edges = sorted(set(sorted_edges), key=sorted_edges.index, reverse=True)
        new_vertex_edges = sorted(set(new_vertex_edges), key=new_vertex_edges.index, reverse=True)
        ref_edge = sorted_edges.index(bary_grid.data().element_edges[vertex_edges[0][1]][vertex_edges[0][0]])
    else:
        # If the first element is not on the border, sorting is not necessary
        # and we remove the barycentric edges that lie on the reference edge.
        vertex_edges.pop(0)
        vertex_edges.pop(-1)

    return vertex_edges, sorted_edges, num_bary_elements // 2, ref_edge


def get_vertex_edges(vertex_index, bary_vertex_to_edge, bary_element, bary_grid):
    """Get all the barycentryc vertices of the elements associated to a reference edge."""
    for ind, elem in enumerate(bary_vertex_to_edge[vertex_index]):
        if bary_element == elem[0]:
            break
    # Get all the relevant edges starting to count above
    # ind
    num_bary_elements = len(bary_vertex_to_edge[vertex_index])
    vertex_edges = []
    for index in range(num_bary_elements):
        elem_edge_pair = bary_vertex_to_edge[vertex_index][(index + ind) % num_bary_elements]
        for n in range(1, 3):
            vertex_edges.append((elem_edge_pair[0], elem_edge_pair[n]))
    edges = []
    for vertex in vertex_edges:
        edges.append(bary_grid.data().element_edges[vertex[1]][vertex[0]])
    set_edges = set(edges)
    sorted_edges = []
    new_vertex_edges = []
    ref_edge = 0

    for el in set_edges:
        if edges.count(el) == 1 and vertex_edges[edges.index(el)][1] == 1:
            sorted_edges.append(el)
            new_vertex_edges.append(vertex_edges[edges.index(el)])

    if len(sorted_edges) > 0:
        new_vertex_edges, sorted_edges = _sort_vertex_edges(
            vertex_edges, edges, set_edges, sorted_edges, new_vertex_edges, 1
        )
        sorted_edges = sorted(set(sorted_edges), key=sorted_edges.index, reverse=True)
        new_vertex_edges = sorted(set(new_vertex_edges), key=new_vertex_edges.index, reverse=True)
        ref_edge = sorted_edges.index(bary_grid.data().element_edges[vertex_edges[0][1]][vertex_edges[0][0]])
    else:
        # We do not want the reference edge part of this list
        vertex_edges.pop(0)
        vertex_edges.pop(-1)

    return vertex_edges, sorted_edges, num_bary_elements // 2, ref_edge


def _sort_vertex_edges(vertex_edges, edges, set_edges, sorted_edges, new_vertex_edges, edge_number):
    """Sort edges anti clock-wise when the basis function lies on the border of a geometry."""
    if edge_number == 1:
        for element in vertex_edges:
            if element[0] == new_vertex_edges[-1][0] and element[1] == 0:
                sorted_edges.append(edges[vertex_edges.index((new_vertex_edges[-1][0], 0))])
                new_vertex_edges.append(element)
    else:
        indexes = [ind for ind, el in enumerate(edges) if el == sorted_edges[-1]]
        if vertex_edges[indexes[0]][1] == 1:
            sorted_edges.append(sorted_edges[-1])
            new_vertex_edges.append(vertex_edges[indexes[0]])
        elif vertex_edges[indexes[1]][1] == 1:
            sorted_edges.append(sorted_edges[-1])
            new_vertex_edges.append(vertex_edges[indexes[1]])
    if len(new_vertex_edges) < len(vertex_edges):
        return _sort_vertex_edges(vertex_edges, edges, set_edges, sorted_edges, new_vertex_edges, 1 - edge_number)
    return new_vertex_edges, sorted_edges


def _get_bary_coefficients(
    edge_lengths,
    vertex_edges1,
    vertex_edges2,
    sorted_edges1,
    sorted_edges2,
    bary_grid,
    local2global,
    nc1,
    nc2,
    global_dof_index,
    ref_edge1,
    ref_edge2,
):
    """Obtain the edge coefficients associated to the barycentric edges of a grid."""
    values = []
    bary_dofs = []
    coarse_dofs = []

    # Check whether any of the poles of the BC basis function is located on the border of a geometry.
    border_edges1 = _check_if_border_edges(vertex_edges1, bary_grid)
    border_edges2 = _check_if_border_edges(vertex_edges2, bary_grid)

    if border_edges1 and not border_edges2:
        aux_values, aux_bary_dofs, aux_coarse_dofs = _border_barycentric_edges_coefficients(
            edge_lengths, vertex_edges1, sorted_edges1, bary_grid, local2global, -1.0, nc1, global_dof_index, ref_edge1
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
        aux_values, aux_bary_dofs, aux_coarse_dofs = _interior_barycentric_edges_coefficients(
            edge_lengths, vertex_edges2, bary_grid, local2global, 1.0, nc2, global_dof_index
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
    elif not border_edges1 and border_edges2:
        aux_values, aux_bary_dofs, aux_coarse_dofs = _interior_barycentric_edges_coefficients(
            edge_lengths, vertex_edges1, bary_grid, local2global, -1.0, nc1, global_dof_index
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
        aux_values, aux_bary_dofs, aux_coarse_dofs = _border_barycentric_edges_coefficients(
            edge_lengths, vertex_edges2, sorted_edges2, bary_grid, local2global, 1.0, nc2, global_dof_index, ref_edge2
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
    elif border_edges1 and border_edges2:
        aux_values, aux_bary_dofs, aux_coarse_dofs = _border_barycentric_edges_coefficients(
            edge_lengths, vertex_edges1, sorted_edges1, bary_grid, local2global, -1.0, nc1, global_dof_index, ref_edge1
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
        aux_values, aux_bary_dofs, aux_coarse_dofs = _border_barycentric_edges_coefficients(
            edge_lengths, vertex_edges2, sorted_edges2, bary_grid, local2global, 1.0, nc2, global_dof_index, ref_edge2
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
    else:
        aux_values, aux_bary_dofs, aux_coarse_dofs = _interior_barycentric_edges_coefficients(
            edge_lengths, vertex_edges1, bary_grid, local2global, -1.0, nc1, global_dof_index
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
        aux_values, aux_bary_dofs, aux_coarse_dofs = _interior_barycentric_edges_coefficients(
            edge_lengths, vertex_edges2, bary_grid, local2global, 1.0, nc2, global_dof_index
        )
        values += aux_values
        bary_dofs += aux_bary_dofs
        coarse_dofs += aux_coarse_dofs
    return values, bary_dofs, coarse_dofs


def _check_if_border_edges(vertex_edges, bary_grid):
    """Check if an edge belongs to the border of the grid."""
    border_edges = False
    for edge in vertex_edges:
        elem_index, local_edge_index = edge[:]
        edge_index = bary_grid.data().element_edges[local_edge_index, elem_index]
        neighbors = bary_grid.edge_neighbors[edge_index]
        if len(neighbors) == 1:
            border_edges = True
            break
    return border_edges


def _border_barycentric_edges_coefficients(
    edge_lengths, vertex_edges, sorted_edges, bary_grid, local2global, sign, nc, global_dof_index, ref_edge
):
    """Calculate barycentric edge coefficients associated to an interior vertex of the grid."""
    values = []
    bary_dofs = []
    coarse_dofs = []
    count = 0
    signs = sign * _np.array([-1.0, 1.0])
    for edge in vertex_edges:
        elem_index, local_edge_index = edge[:]
        edge_length = edge_lengths[bary_grid.data().element_edges[local_edge_index, elem_index]]
        count = sorted_edges.index(bary_grid.data().element_edges[local_edge_index, elem_index])

        bary_dofs.append(local2global[elem_index, local_edge_index])
        coarse_dofs.append(global_dof_index)
        if count < ref_edge:
            values.append(signs[local_edge_index] * (1 - nc) / (nc * edge_length))
        elif count == ref_edge:
            values.append(signs[local_edge_index] * (2 - nc) / (2 * nc * edge_length))
        else:
            values.append(signs[local_edge_index] * 1.0 / (nc * edge_length))
    return values, bary_dofs, coarse_dofs


def _interior_barycentric_edges_coefficients(
    edge_lengths, vertex_edges, bary_grid, local2global, sign, nc, global_dof_index
):
    """Calculate barycentric edge coefficients associated to a vertex that belongs to the border of the grid.

    (eg: vertex on the edge of a screen).
    """
    values = []
    bary_dofs = []
    coarse_dofs = []
    count = 0
    for index, edge in enumerate(vertex_edges):
        if index % 2 == 0:
            count += 1
        elem_index, local_edge_index = edge[:]
        edge_length = edge_lengths[bary_grid.data().element_edges[local_edge_index, elem_index]]
        bary_dofs.append(local2global[elem_index, local_edge_index])
        coarse_dofs.append(global_dof_index)
        values.append(sign * (nc - count) / (2 * nc * edge_length))
        sign *= -1
    return values, bary_dofs, coarse_dofs


def _get_coefficients_reference_edge(
    edge_lengths,
    bary_grid,
    local2global,
    global_dof_index,
    bary_upper_minus,
    bary_upper_plus,
    bary_lower_minus,
    bary_lower_plus,
):
    """Calculate upper and lower coefficients of the barycentric edges associated to a reference edge."""
    values = []
    bary_dofs = []
    coarse_dofs = []

    coarse_dofs.append(global_dof_index)
    coarse_dofs.append(global_dof_index)
    coarse_dofs.append(global_dof_index)
    coarse_dofs.append(global_dof_index)

    edge_length_upper = edge_lengths[bary_grid.data().element_edges[2, bary_upper_minus]]
    edge_length_lower = edge_lengths[bary_grid.data().element_edges[2, bary_lower_minus]]

    bary_dofs.append(local2global[bary_upper_minus, 2])
    bary_dofs.append(local2global[bary_upper_plus, 2])
    bary_dofs.append(local2global[bary_lower_minus, 2])
    bary_dofs.append(local2global[bary_lower_plus, 2])

    values.append(1.0 / (2 * edge_length_upper))
    values.append(-1.0 / (2 * edge_length_upper))
    values.append(-1.0 / (2 * edge_length_lower))
    values.append(1.0 / (2 * edge_length_lower))

    return values, bary_dofs, coarse_dofs
